# pylint: disable=wrong-import-position
# pylint: disable=too-many-lines
#
# Copyright (C) 2009 Mathieu Comandon <mathieucomandon@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import logging
import os
import signal
import sys
import tempfile
from datetime import datetime, timedelta
from gettext import gettext as _
from typing import List, Optional

import gi

gi.require_version("Gdk", "3.0")
gi.require_version("Gtk", "3.0")

from gi.repository import Gio, GLib, Gtk

from lutris import settings
from lutris.api import get_runners, parse_installer_url
from lutris.database import games as games_db
from lutris.database.services import ServiceGameCollection
from lutris.exception_backstops import init_exception_backstops
from lutris.game import GAME_START, GAME_STOPPED, Game, export_game, import_game
from lutris.gui.dialogs import ErrorDialog, InstallOrPlayDialog, NoticeDialog, display_error
from lutris.gui.dialogs.delegates import CommandLineUIDelegate, InstallUIDelegate, LaunchUIDelegate
from lutris.gui.dialogs.issue import IssueReportWindow
from lutris.gui.installerwindow import INSTALLATION_COMPLETED, INSTALLATION_FAILED, InstallationKind, InstallerWindow
from lutris.gui.widgets.status_icon import LutrisStatusIcon
from lutris.installer import get_installers
from lutris.migrations import migrate
from lutris.monitored_command import exec_command
from lutris.runners import InvalidRunnerError, RunnerInstallationError, get_runner_names, import_runner
from lutris.services import get_enabled_services
from lutris.startup import init_lutris, run_all_checks
from lutris.style_manager import StyleManager
from lutris.util import datapath, log, resources, system
from lutris.util.http import HTTPError, Request
from lutris.util.log import file_handler, logger
from lutris.util.savesync import save_check, show_save_stats, upload_save
from lutris.util.steam.appmanifest import AppManifest, get_appmanifests
from lutris.util.steam.config import get_steamapps_dirs

from ..util.busy import BusyAsyncCall
from ..util.standalone_scripts import generate_script
from .lutriswindow import LutrisWindow

LUTRIS_EXPERIMENTAL_FEATURES_ENABLED = os.environ.get("LUTRIS_EXPERIMENTAL_FEATURES_ENABLED") == "1"


class LutrisApplication(Gtk.Application):
    def __init__(self) -> None:
        super().__init__(
            application_id="net.lutris.Lutris",
            flags=Gio.ApplicationFlags.HANDLES_COMMAND_LINE,
            register_session=True,
        )

        # Prepare the backstop logic just before the first emission hook (or connection) is
        # established; this will apply to all connections from this point forward.
        init_exception_backstops()

        GAME_START.register(self.on_game_start)
        GAME_STOPPED.register(self.on_game_stopped)
        settings.SETTINGS_CHANGED.register(self.on_settings_changed)
        INSTALLATION_COMPLETED.register(self.on_install_ended)
        INSTALLATION_FAILED.register(self.on_install_ended)

        GLib.set_application_name(_("Lutris"))
        GLib.set_prgname("net.lutris.Lutris")
        self.force_updates = False
        self.css_provider = Gtk.CssProvider.new()
        self.window: Optional[LutrisWindow] = None
        self.launch_ui_delegate = LaunchUIDelegate()
        self.install_ui_delegate = InstallUIDelegate()

        self._running_games = []
        self.app_windows = {}
        self.tray = None

        self.quit_on_game_exit = False
        self.style_manager = StyleManager()

        if os.geteuid() == 0:
            NoticeDialog(_("Do not run Lutris as root."))
            sys.exit(2)

        try:
            self.css_provider.load_from_path(os.path.join(datapath.get(), "ui", "lutris.css"))
        except GLib.Error as e:
            logger.exception(e)

        if hasattr(self, "add_main_option"):
            self.add_arguments()
        else:
            ErrorDialog(_("Your Linux distribution is too old. Lutris won't function properly."))

    def add_arguments(self):
        if hasattr(self, "set_option_context_summary"):
            self.set_option_context_summary(
                _(
                    "Run a game directly by adding the parameter lutris:rungame/game-identifier.\n"
                    "If several games share the same identifier you can use the numerical ID "
                    "(displayed when running lutris --list-games) and add "
                    "lutris:rungameid/numerical-id.\n"
                    "To install a game, add lutris:install/game-identifier."
                )
            )
        else:
            logger.warning("GLib.set_option_context_summary missing, was added in GLib 2.56 (Released 2018-03-12)")
        self.add_main_option(
            "version",
            ord("v"),
            GLib.OptionFlags.NONE,
            GLib.OptionArg.NONE,
            _("Print the version of Lutris and exit"),
            None,
        )
        self.add_main_option(
            "debug",
            ord("d"),
            GLib.OptionFlags.NONE,
            GLib.OptionArg.NONE,
            _("Show debug messages"),
            None,
        )
        self.add_main_option(
            "install",
            ord("i"),
            GLib.OptionFlags.NONE,
            GLib.OptionArg.STRING,
            _("Install a game from a yml file"),
            None,
        )
        self.add_main_option(
            "force",
            ord("f"),
            GLib.OptionFlags.NONE,
            GLib.OptionArg.NONE,
            _("Force updates"),
            None,
        )
        self.add_main_option(
            "output-script",
            ord("b"),
            GLib.OptionFlags.NONE,
            GLib.OptionArg.STRING,
            _("Generate a bash script to run a game without the client"),
            None,
        )
        self.add_main_option(
            "exec",
            ord("e"),
            GLib.OptionFlags.NONE,
            GLib.OptionArg.STRING,
            _("Execute a program with the Lutris Runtime"),
            None,
        )
        self.add_main_option(
            "list-games",
            ord("l"),
            GLib.OptionFlags.NONE,
            GLib.OptionArg.NONE,
            _("List all games in database"),
            None,
        )
        self.add_main_option(
            "installed",
            ord("o"),
            GLib.OptionFlags.NONE,
            GLib.OptionArg.NONE,
            _("Only list installed games"),
            None,
        )
        self.add_main_option(
            "list-steam-games",
            ord("s"),
            GLib.OptionFlags.NONE,
            GLib.OptionArg.NONE,
            _("List available Steam games"),
            None,
        )
        self.add_main_option(
            "list-steam-folders",
            0,
            GLib.OptionFlags.NONE,
            GLib.OptionArg.NONE,
            _("List all known Steam library folders"),
            None,
        )
        self.add_main_option(
            "list-runners",
            0,
            GLib.OptionFlags.NONE,
            GLib.OptionArg.NONE,
            _("List all known runners"),
            None,
        )
        self.add_main_option(
            "list-wine-versions",
            0,
            GLib.OptionFlags.NONE,
            GLib.OptionArg.NONE,
            _("List all known Wine versions"),
            None,
        )
        self.add_main_option(
            "list-all-service-games",
            ord("a"),
            GLib.OptionFlags.NONE,
            GLib.OptionArg.NONE,
            _("List all games for all services in database"),
            None,
        )
        self.add_main_option(
            "list-service-games",
            0,
            GLib.OptionFlags.NONE,
            GLib.OptionArg.STRING,
            _("List all games for provided service in database"),
            None,
        )
        self.add_main_option(
            "install-runner",
            ord("r"),
            GLib.OptionFlags.NONE,
            GLib.OptionArg.STRING,
            _("Install a Runner"),
            None,
        )
        self.add_main_option(
            "uninstall-runner",
            ord("u"),
            GLib.OptionFlags.NONE,
            GLib.OptionArg.STRING,
            _("Uninstall a Runner"),
            None,
        )
        self.add_main_option(
            "export",
            0,
            GLib.OptionFlags.NONE,
            GLib.OptionArg.STRING,
            _("Export a game"),
            None,
        )
        self.add_main_option(
            "import",
            0,
            GLib.OptionFlags.NONE,
            GLib.OptionArg.STRING,
            _("Import a game"),
            None,
        )
        if LUTRIS_EXPERIMENTAL_FEATURES_ENABLED:
            self.add_main_option(
                "save-stats",
                0,
                GLib.OptionFlags.NONE,
                GLib.OptionArg.STRING,
                _("Show statistics about a game saves"),
                None,
            )
            self.add_main_option(
                "save-upload",
                0,
                GLib.OptionFlags.NONE,
                GLib.OptionArg.STRING,
                _("Upload saves"),
                None,
            )
            self.add_main_option(
                "save-check",
                0,
                GLib.OptionFlags.NONE,
                GLib.OptionArg.STRING,
                _("Verify status of save syncing"),
                None,
            )

        self.add_main_option(
            "dest",
            0,
            GLib.OptionFlags.NONE,
            GLib.OptionArg.STRING,
            _("Destination path for export"),
            None,
        )
        self.add_main_option(
            "json",
            ord("j"),
            GLib.OptionFlags.NONE,
            GLib.OptionArg.NONE,
            _("Display the list of games in JSON format"),
            None,
        )
        self.add_main_option(
            "reinstall",
            0,
            GLib.OptionFlags.NONE,
            GLib.OptionArg.NONE,
            _("Reinstall game"),
            None,
        )
        self.add_main_option("submit-issue", 0, GLib.OptionFlags.NONE, GLib.OptionArg.NONE, _("Submit an issue"), None)
        self.add_main_option(
            GLib.OPTION_REMAINING,
            0,
            GLib.OptionFlags.NONE,
            GLib.OptionArg.STRING_ARRAY,
            _("URI to open"),
            "URI",
        )

    def do_startup(self):  # pylint: disable=arguments-differ
        """Sets up the application on first start."""
        Gtk.Application.do_startup(self)
        file_handler.doRollover()
        signal.signal(signal.SIGINT, signal.SIG_DFL)

        action = Gio.SimpleAction.new("quit")
        action.connect("activate", lambda *x: self.quit())
        self.add_action(action)
        self.add_accelerator("<Primary>q", "app.quit")

    def do_activate(self):  # pylint: disable=arguments-differ
        if not self.window:
            self.window = LutrisWindow(application=self)
            screen = self.window.props.screen  # pylint: disable=no-member
            Gtk.StyleContext.add_provider_for_screen(screen, self.css_provider, Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION)

    def start_runtime_updates(self) -> None:
        if os.environ.get("LUTRIS_SKIP_INIT"):
            logger.debug("Skipping initialization")
        elif not self.window:
            logger.error("Runtime updates can't be started with a LutrisWindow.")
        else:
            self.window.start_runtime_updates(self.force_updates)

    def get_window_key(self, **kwargs):
        if kwargs.get("appid"):
            return kwargs["appid"]
        if kwargs.get("runner"):
            return kwargs["runner"].name
        if kwargs.get("installers"):
            installer = kwargs["installers"][0]
            return installer.get("slug") or installer.get("game_slug")
        if kwargs.get("game"):
            return kwargs["game"].id
        return str(kwargs)

    def show_window(self, window_class, update_function=None, **kwargs):
        """Instantiate a window keeping 1 instance max

        Params:
            window_class (Gtk.Window): class to create the instance from
            update_function (Callable): Function to initialize or update the window (if possible before being shown)
            kwargs (dict): Additional arguments to pass to the instanciated window

        Returns:
            Gtk.Window: the existing window instance or a newly created one
        """
        window_key = str(window_class.__name__) + self.get_window_key(**kwargs)
        if self.app_windows.get(window_key):
            self.app_windows[window_key].present()
            window_inst = self.app_windows[window_key]
            if update_function:
                update_function(window_inst)
            return window_inst
        if issubclass(window_class, Gtk.Dialog):
            if "parent" in kwargs:
                window_inst = window_class(**kwargs)
            else:
                window_inst = window_class(parent=self.window, **kwargs)
            window_inst.set_application(self)
        else:
            window_inst = window_class(application=self, **kwargs)
        window_inst.connect("destroy", self.on_app_window_destroyed, self.get_window_key(**kwargs))
        if update_function:
            update_function(window_inst)
        self.app_windows[window_key] = window_inst
        window_inst.show()
        window_inst.present()
        return window_inst

    def show_installer_window(self, installers, service=None, appid=None, installation_kind=InstallationKind.INSTALL):
        self.show_window(
            InstallerWindow, installers=installers, service=service, appid=appid, installation_kind=installation_kind
        )

    def show_lutris_installer_window(self, game_slug):
        def on_installers_ready(installers, error):
            if error:
                display_error(error, parent=self.window)
            elif installers:
                self.show_installer_window(installers)
            else:
                ErrorDialog(_("No installer available."), parent=self.window)

        BusyAsyncCall(get_installers, on_installers_ready, game_slug=game_slug)

    def on_app_window_destroyed(self, app_window, window_key):
        """Remove the reference to the window when it has been destroyed"""
        window_key = str(app_window.__class__.__name__) + window_key
        try:
            del self.app_windows[window_key]
        except KeyError:
            logger.warning("Failed to remove window %s", window_key)
            logger.info("Available windows: %s", ", ".join(self.app_windows.keys()))
        return True

    @staticmethod
    def _print(command_line, string):
        # Workaround broken pygobject bindings
        command_line.do_print_literal(command_line, string + "\n")

    def do_handle_local_options(self, options):
        # Text only commands

        # Print Lutris version and exit
        if options.contains("version"):
            executable_name = os.path.basename(sys.argv[0])
            print(executable_name + "-" + settings.VERSION)
            logger.setLevel(logging.NOTSET)
            return 0
        return -1  # continue command line processes

    def do_command_line(self, command_line):  # noqa: C901  # pylint: disable=arguments-differ
        # pylint: disable=too-many-locals,too-many-return-statements,too-many-branches
        # pylint: disable=too-many-statements
        options = command_line.get_options_dict()

        # Use stdout to output logs, only if no command line argument is
        # provided.
        argc = len(sys.argv) - 1
        if "-d" in sys.argv or "--debug" in sys.argv:
            argc -= 1
        if not argc:
            # Switch back the log output to stderr (the default in Python)
            # to avoid messing with any output from command line options.
            log.console_handler.setStream(sys.stderr)

        # Set up logger
        if options.contains("debug"):
            log.console_handler.setFormatter(log.DEBUG_FORMATTER)
            logger.setLevel(logging.DEBUG)

        if options.contains("force"):
            self.force_updates = True

        logger.info("Starting Lutris %s", settings.VERSION)
        init_lutris()
        migrate()

        run_all_checks()
        if options.contains("dest"):
            dest_dir = options.lookup_value("dest").get_string()
        else:
            dest_dir = None

        if options.contains("output-script"):
            searchstring = options.lookup_value("output-script").get_string()
            export_script_game = (
                games_db.get_game_by_field(searchstring, "id")
                or games_db.get_game_by_field(searchstring, "slug")
                or games_db.get_game_by_field(searchstring, "installer_slug")
            )

            if not export_script_game or not export_script_game["id"]:
                logger.warning(
                    "No valid game (%s) provided to generate the script. Use the -l option to find suitable games!",
                    searchstring,
                )
                return 1
            generate_script(logger, self.launch_ui_delegate, export_script_game, f"{export_script_game['slug']}.sh")
            return 0

        # List game
        if options.contains("list-games"):
            game_list = games_db.get_games(filters=({"installed": 1} if options.contains("installed") else None))
            if options.contains("json"):
                self.print_game_json(command_line, game_list)
            else:
                self.print_game_list(command_line, game_list)
            return 0

        # List specified service games
        if options.contains("list-service-games"):
            service = options.lookup_value("list-service-games").get_string()
            game_list = games_db.get_games(filters={"installed": 1, "service": service})
            service_game_list = ServiceGameCollection.get_for_service(service)
            for game in service_game_list:
                game["installed"] = any(("service_id", game["appid"]) in item.items() for item in game_list)
            if options.contains("installed"):
                service_game_list = [d for d in service_game_list if d["installed"]]
            if options.contains("json"):
                self.print_service_game_json(command_line, service_game_list)
            else:
                self.print_service_game_list(command_line, service_game_list)
            return 0

        # List all service games
        if options.contains("list-all-service-games"):
            game_list = games_db.get_games(filters={"installed": 1})
            service_game_list = ServiceGameCollection.get_service_games()
            for game in service_game_list:
                game["installed"] = any(
                    ("service_id", game["appid"]) in item.items()
                    for item in game_list
                    if item["service"] == game["service"]
                )
            if options.contains("installed"):
                service_game_list = [d for d in service_game_list if d["installed"]]
            if options.contains("json"):
                self.print_service_game_json(command_line, service_game_list)
            else:
                self.print_service_game_list(command_line, service_game_list)
            return 0

        # List Steam games
        if options.contains("list-steam-games"):
            self.print_steam_list(command_line)
            return 0

        # List Steam folders
        if options.contains("list-steam-folders"):
            self.print_steam_folders(command_line)
            return 0

        # List Runners
        if options.contains("list-runners"):
            self.print_runners()
            return 0

        # List Wine Runners
        if options.contains("list-wine-versions"):
            self.print_wine_runners()
            return 0

        # install Runner
        if options.contains("install-runner"):
            runner = options.lookup_value("install-runner").get_string()
            self.install_runner(runner)
            return 0

        # Uninstall Runner
        if options.contains("uninstall-runner"):
            runner = options.lookup_value("uninstall-runner").get_string()
            self.uninstall_runner(runner)
            return 0

        if options.contains("export"):
            slug = options.lookup_value("export").get_string()
            if not dest_dir:
                print("No destination dir given")
            else:
                export_game(slug, dest_dir)
            return 0

        if options.contains("import"):
            filepath = options.lookup_value("import").get_string()
            if not dest_dir:
                print("No destination dir given")
            else:
                import_game(filepath, dest_dir)
            return 0

        if LUTRIS_EXPERIMENTAL_FEATURES_ENABLED:

            def get_game_match(slug):
                # First look for an exact match
                games = games_db.get_games_by_slug(slug)
                if not games:
                    # Then look for matches
                    games = games_db.get_games(searches={"slug": slug})
                if len(games) > 1:
                    self._print(
                        command_line,
                        "Multiple games matching %s: %s" % (slug, ",".join(game["slug"] for game in games)),
                    )
                    return
                if not games:
                    self._print(command_line, "No matching game for %s" % slug)
                    return
                return Game(games[0]["id"])

            if options.contains("save-stats"):
                game = get_game_match(options.lookup_value("save-stats").get_string())
                if game:
                    show_save_stats(game, output_format="json" if options.contains("json") else "text")
                return 0
            if options.contains("save-upload"):
                game = get_game_match(options.lookup_value("save-upload").get_string())
                if game:
                    upload_save(game)
                return 0
            if options.contains("save-check"):
                game = get_game_match(options.lookup_value("save-check").get_string())
                if game:
                    save_check(game)
                return 0

        # Execute command in Lutris context
        if options.contains("exec"):
            command = options.lookup_value("exec").get_string()
            self.execute_command(command)
            return 0

        if options.contains("submit-issue"):
            IssueReportWindow(application=self)
            return 0

        url = options.lookup_value(GLib.OPTION_REMAINING)
        try:
            installer_info = self.get_lutris_action(url)
        except ValueError:
            self._print(command_line, _("%s is not a valid URI") % url.get_strv())
            return 1

        game_slug = installer_info["game_slug"]
        action = installer_info["action"]
        service = installer_info["service"]
        appid = installer_info["appid"]
        launch_config_name = installer_info["launch_config_name"]

        self.launch_ui_delegate = CommandLineUIDelegate(launch_config_name)

        revision = installer_info["revision"]

        installer_file = None
        if options.contains("install"):
            installer_file = options.lookup_value("install").get_string()
            if installer_file.startswith(("http:", "https:")):
                try:
                    request = Request(installer_file).get()
                except HTTPError:
                    self._print(command_line, _("Failed to download %s") % installer_file)
                    return 1
                try:
                    headers = dict(request.response_headers)
                    file_name = headers["Content-Disposition"].split("=", 1)[-1]
                except (KeyError, IndexError):
                    file_name = os.path.basename(installer_file)
                file_path = os.path.join(tempfile.gettempdir(), file_name)
                self._print(
                    command_line, _("download {url} to {file} started").format(url=installer_file, file=file_path)
                )
                with open(file_path, "wb") as dest_file:
                    dest_file.write(request.content)
                installer_file = file_path
                action = "install"
            else:
                # Use the invoking process's cwd for relative paths, not the running instance's
                if not os.path.isabs(installer_file):
                    cwd = command_line.get_cwd()
                    if cwd:
                        installer_file = os.path.join(cwd, installer_file)
                installer_file = os.path.abspath(installer_file)
                action = "install"

            if not os.path.isfile(installer_file):
                self._print(command_line, _("No such file: %s") % installer_file)
                return 1

        db_game = None
        if game_slug and not service:
            if action == "rungameid":
                # Force db_game to use game id
                self.quit_on_game_exit = True
                db_game = games_db.get_game_by_field(game_slug, "id")
            elif action == "rungame":
                # Force db_game to use game slug
                self.quit_on_game_exit = True
                db_game = games_db.get_game_by_field(game_slug, "slug")
            elif action == "install":
                # Installers can use game or installer slugs
                self.quit_on_game_exit = True
                db_game = games_db.get_game_by_field(game_slug, "slug") or games_db.get_game_by_field(
                    game_slug, "installer_slug"
                )
            else:
                # Dazed and confused, try anything that might works
                db_game = (
                    games_db.get_game_by_field(game_slug, "id")
                    or games_db.get_game_by_field(game_slug, "slug")
                    or games_db.get_game_by_field(game_slug, "installer_slug")
                )

        # If reinstall flag is passed, force the action to install
        if options.contains("reinstall"):
            action = "install"

        # Graphical commands
        self.set_tray_icon()
        self.activate()

        if not action:
            if db_game and db_game["installed"]:
                # Game found but no action provided, ask what to do
                dlg = InstallOrPlayDialog(db_game["name"])
                if not dlg.action:
                    action = "cancel"
                elif dlg.action == "play":
                    action = "rungame"
                elif dlg.action == "install":
                    action = "install"
            elif game_slug or installer_file or service:
                # No game found, default to install if a game_slug or
                # installer_file is provided
                action = "install"

        if service:
            service_game = ServiceGameCollection.get_game(service, appid)
            if service_game:
                service = get_enabled_services()[service]()
                service.install(service_game)
                return 0

        if action == "cancel":
            if not self.window.is_visible():
                self.quit()
            return 0

        if action == "install":
            installers = get_installers(
                game_slug=game_slug,
                installer_file=installer_file,
                revision=revision,
            )
            if installers:
                self.show_installer_window(installers)
        elif action in ("rungame", "rungameid"):
            if not db_game or not db_game["id"]:
                logger.warning("No game found in library")
                if not self.window.is_visible():
                    self.quit()
                return 0

            def on_error(error: BaseException) -> None:
                logger.exception("Unable to launch game: %s", error)

            game = Game(db_game["id"])
            game.game_error.register(on_error)
            game.launch(self.launch_ui_delegate)

            if game.state == game.STATE_STOPPED and not self.window.is_visible():
                self.quit()

            if self.quit_on_game_exit:

                def game_stop_signal_handler(signum, _frame):
                    logger.debug("signal handler called with signal: %d", signum)
                    game.stop()

                signal.signal(signal.SIGTERM, game_stop_signal_handler)
                signal.signal(signal.SIGINT, game_stop_signal_handler)
        elif self.window:
            # If we're showing the window, it will handle the delegated UI
            # from here on out, no matter what command line we got.
            self.launch_ui_delegate = self.window
            self.install_ui_delegate = self.window
            self.window.present()
            self.start_runtime_updates()
            # If the Lutris GUI is started by itself, don't quit it when a game stops
            self.quit_on_game_exit = False
        return 0

    def on_settings_changed(self, setting_key, new_value, section):
        if section == "lutris":
            if setting_key == "preferred_theme":
                self.style_manager.preferred_theme = new_value
            elif setting_key == "show_tray_icon" and self.window:
                if self.window.get_visible():
                    self.set_tray_icon()
        return True

    def on_game_start(self, game: Game) -> None:
        self._running_games.append(game)
        if self.window and settings.read_bool_setting("hide_client_on_game_start"):
            self.window.hide()  # Hide launcher window

    def on_game_stopped(self, game: Game) -> None:
        """Callback to quit Lutris is last game stops while the window is hidden."""
        running_game_ids = [g.id for g in self._running_games]
        if game.id in running_game_ids:
            logger.debug("Removing %s from running IDs", game.id)
            try:
                del self._running_games[running_game_ids.index(game.id)]
            except ValueError:
                pass
        elif running_game_ids:
            logger.warning("%s not in %s", game.id, running_game_ids)
        else:
            logger.debug("Game has already been removed from running IDs?")

        if self.window and settings.read_bool_setting("hide_client_on_game_start") and not self.quit_on_game_exit:
            self.window.show()  # Show launcher window
        elif not self.window or not self.window.is_visible():
            if not self.has_running_games:
                if self.quit_on_game_exit or not self.has_tray_icon():
                    self.quit()

    def on_install_ended(self):
        if not self.window or not self.window.is_visible():
            if not self.has_running_games:
                if self.quit_on_game_exit or not self.has_tray_icon():
                    self.quit()

    def get_launch_ui_delegate(self):
        return self.launch_ui_delegate

    def get_running_games(self) -> List[Game]:
        # This method reflects games that have stopped even if the 'game-stopped' signal
        # has not been handled yet; that handler will still clean up the list though.
        return [g for g in self._running_games if g.state != g.STATE_STOPPED]

    @property
    def has_running_games(self):
        return bool(self.get_running_games())

    def get_running_game_ids(self) -> List[str]:
        """Returns the ids of the games presently running."""
        return [game.id for game in self.get_running_games()]

    def is_game_running_by_id(self, game_id: str) -> bool:
        """True if the ID is the ID of a game that is running."""
        return bool(game_id and str(game_id) in self.get_running_game_ids())

    def get_game_by_id(self, game_id: str) -> Game:
        """Returns the game with the ID given; if it's running this is the running
        game instance, and if not it's a fresh copy off the database."""
        for game in self.get_running_games():
            if game.id == str(game_id):
                return game

        return Game(game_id)

    @staticmethod
    def get_lutris_action(url):
        installer_info = {
            "game_slug": None,
            "revision": None,
            "action": None,
            "service": None,
            "appid": None,
            "launch_config_name": None,
        }

        if url:
            url = url.get_strv()

        if url:
            url = url[0]
            installer_info = parse_installer_url(url)
        return installer_info

    def print_game_list(self, command_line, game_list):
        for game in game_list:
            self._print(
                command_line,
                "{:4} | {:<40} | {:<40} | {:<15} | {:<64}".format(
                    game["id"],
                    game["name"][:40],
                    game["slug"][:40],
                    game["runner"] or "-",
                    game["directory"] or "-",
                ),
            )

    def print_game_json(self, command_line, game_list):
        games = []

        for game in game_list:
            playtime = timedelta(hours=game["playtime"]) if game["playtime"] else None
            cover_path = resources.get_cover_path(game["slug"]) if game["slug"] else None

            if cover_path and not os.path.exists(cover_path):
                cover_path = None

            game_obj = {
                "id": game["id"],
                "slug": game["slug"],
                "name": game["name"],
                "runner": game["runner"],
                "platform": game["platform"] or None,
                "year": game["year"] or None,
                "directory": game["directory"] or None,
                "playtime": str(playtime) if playtime else None,
                "playtimeSeconds": playtime.total_seconds() if playtime else None,
                "lastplayed": (str(datetime.fromtimestamp(game["lastplayed"])) if game["lastplayed"] else None),
                "coverPath": cover_path,
            }

            games.append(game_obj)

        self._print(command_line, json.dumps(games, indent=2))

    def print_service_game_list(self, command_line, game_list):
        for game in game_list:
            self._print(
                command_line,
                "{:4} | {:<40} | {:<40} | {:<15} | {:<15}".format(
                    game["id"],
                    game["name"][:40],
                    game["slug"][:40],
                    game["service"][:15],
                    "true" if game["installed"] else "false"[:15],
                ),
            )

    def print_service_game_json(self, command_line, game_list):
        games = [
            {
                "id": game["id"],
                "slug": game["slug"],
                "name": game["name"],
                "service": game["service"],
                "appid": game["appid"],
                "installed": game["installed"],
                "details": game["details"],
            }
            for game in game_list
        ]
        self._print(command_line, json.dumps(games, indent=2))

    def print_steam_list(self, command_line):
        steamapps_paths = get_steamapps_dirs()
        for path in steamapps_paths if steamapps_paths else []:
            appmanifest_files = get_appmanifests(path)
            for appmanifest_file in appmanifest_files:
                appmanifest = AppManifest(os.path.join(path, appmanifest_file))
                self._print(
                    command_line,
                    " {:8} | {:<60} | {}".format(
                        appmanifest.steamid,
                        appmanifest.name or "-",
                        ", ".join(appmanifest.states),
                    ),
                )

    @staticmethod
    def execute_command(command):
        """Execute an arbitrary command in a Lutris context
        with the runtime enabled and monitored by a MonitoredCommand
        """
        logger.info("Running command '%s'", command)
        monitored_command = exec_command(command)
        try:
            GLib.MainLoop().run()
        except KeyboardInterrupt:
            monitored_command.stop()

    def print_steam_folders(self, command_line):
        steamapps_paths = get_steamapps_dirs()
        if steamapps_paths:
            for path in steamapps_paths:
                self._print(command_line, path)

    def print_runners(self):
        runner_names = get_runner_names()
        sorted_names = sorted(runner_names, key=lambda x: x.lower())
        for name in sorted_names:
            print(name)

    def print_wine_runners(self):
        runnersName = get_runners("wine")
        for i in runnersName["versions"]:
            if i["version"]:
                print(i)

    def install_runner(self, runner):
        if runner.startswith("lutris"):
            self.install_wine_cli(runner)
        else:
            self.install_cli(runner)

    def uninstall_runner(self, runner):
        if "wine" in runner:
            print("Are sure you want to delete Wine and all of the installed runners?[Y/N]")
            ans = input()
            if ans.lower() in ("y", "yes"):
                self.uninstall_runner_cli(runner)
            else:
                print("Not Removing Wine")
        elif runner.startswith("lutris"):
            self.wine_runner_uninstall(runner)
        else:
            self.uninstall_runner_cli(runner)

    def install_wine_cli(self, version):
        """
        Downloads wine runner using lutris -r <runner>
        """

        WINE_DIR = os.path.join(settings.RUNNER_DIR, "wine")
        runner_path = os.path.join(WINE_DIR, f"{version}{'' if '-x86_64' in version else '-x86_64'}")
        if os.path.isdir(runner_path):
            print(f"Wine version '{version}' is already installed.")
        else:
            try:
                runner = import_runner("wine")
                runner().install(self.install_ui_delegate, version=version)
                print(f"Wine version '{version}' has been installed.")
            except (InvalidRunnerError, RunnerInstallationError) as ex:
                print(ex.message)

    def wine_runner_uninstall(self, version):
        version = f"{version}{'' if '-x86_64' in version else '-x86_64'}"
        WINE_DIR = os.path.join(settings.RUNNER_DIR, "wine")
        runner_path = os.path.join(WINE_DIR, version)
        if os.path.isdir(runner_path):
            system.remove_folder(runner_path)
            print(f"Wine version '{version}' has been removed.")
        else:
            print(
                f"""
Specified version of Wine is not installed: {version}.
Please check if the Wine Runner and specified version are installed (for that use --list-wine-versions).
Also, check that the version specified is in the correct format.
                """
            )

    def install_cli(self, runner_name):
        """
        install the runner provided in prepare_runner_cli()
        """

        try:
            runner = import_runner(runner_name)()
            if runner.is_installed():
                print(f"'{runner_name}' is already installed.")
            else:
                runner.install(self.install_ui_delegate, version=None, callback=None)
                print(f"'{runner_name}' has been installed")
        except (InvalidRunnerError, RunnerInstallationError) as ex:
            print(ex.message)

    def uninstall_runner_cli(self, runner_name):
        """
        uninstall the runner given in application file located in lutris/gui/application.py
        provided using lutris -u <runner>
        """
        try:
            runner_class = import_runner(runner_name)
            runner = runner_class()
        except InvalidRunnerError:
            logger.error("Failed to import Runner: %s", runner_name)
            return
        if not runner.is_installed():
            print(f"Runner '{runner_name}' is not installed.")
            return
        if runner.can_uninstall():
            runner.uninstall()
            print(f"'{runner_name}' has been uninstalled.")
        else:
            print(f"Runner '{runner_name}' cannot be uninstalled.")

    def do_shutdown(self):  # pylint: disable=arguments-differ
        logger.info("Shutting down Lutris")
        if self.window:
            selected_category = "%s:%s" % self.window.selected_category
            settings.write_setting("selected_category", selected_category)
            self.window.destroy()
        Gtk.Application.do_shutdown(self)

    def set_tray_icon(self):
        """Creates or destroys a tray icon for the application"""
        active = settings.read_setting("show_tray_icon", default="false").lower() == "true"
        if active and not self.tray:
            self.tray = LutrisStatusIcon(application=self)
        if self.tray:
            self.tray.set_visible(active)

    def has_tray_icon(self):
        return self.tray and self.tray.is_visible()



import shutil
from gettext import gettext as _
from typing import Dict, List, Optional

from lutris.exceptions import UnavailableRunnerError
from lutris.util import cache_single
from lutris.util.log import logger
from lutris.util.system import read_process_output


@cache_single
def get_executable() -> Optional[str]:
    """Return the executable used to access Flatpak. None if Flatpak is not installed.

    In the case where Lutris is a Flatpak, we use flatpak-spawn.
    """
    return shutil.which("flatpak-spawn") or shutil.which("flatpak")


def is_installed() -> bool:
    """Returns Flatpak is installed"""
    return bool(get_executable())


def get_command() -> List[str]:
    """Return the full command used to interact with Flatpak."""
    exe = get_executable()
    if not exe:
        raise UnavailableRunnerError(_("Flatpak is not installed"))
    if "flatpak-spawn" in exe:
        return [exe, "--host", "flatpak"]
    return [exe]


@cache_single
def get_installed_apps() -> List[Dict[str, str]]:
    if not is_installed():
        return []

    command = get_command() + ["list", "--columns=name,application,version,branch,arch,origin,installation"]
    package_list = read_process_output(command)
    packages = []
    for package in package_list.split("\n"):
        if package:
            try:
                name, appid, version, branch, arch, origin, installation = package.split("\t")
            except ValueError:
                logger.error("Not able to parse Flatpak output: %s", package)
                continue
            packages.append(
                {
                    "name": name,
                    "appid": appid,
                    "version": version,
                    "branch": branch,
                    "arch": arch,
                    "origin": origin,
                    "installation": installation,
                }
            )
    return packages


def is_app_installed(appid: str) -> bool:
    """Return whether an app is installed"""
    if not appid:
        return False
    for app in get_installed_apps():
        if app["appid"] == appid:
            return True
    return False


def get_run_command(appid: str, arch: str = None, fcommand: str = None, branch: str = None) -> List[str]:
    """Return command to launch a Flatpak app"""
    command = get_bare_run_command(arch, fcommand, branch)
    command.append(appid)
    return command


def get_bare_run_command(arch: str = None, fcommand: str = None, branch: str = None) -> List[str]:
    """Return command to launch a Flatpak app, without the app-id at the end;
    this is the 'command' of the flatpak runner itself, and a program's
    appid must be appended to it."""
    command = get_command()
    command.append("run")
    if arch:
        command.append(f"--arch={arch}")
    if fcommand:
        command.append(f"--command={fcommand}")
    if branch:
        command.append(f"--branch={branch}")
    return command



"""Runner for Linux games"""

# Standard Library
import os
import stat
from gettext import gettext as _
from typing import Callable

# Lutris Modules
from lutris.exceptions import GameConfigError, MissingGameExecutableError
from lutris.runners.runner import Runner
from lutris.util import system
from lutris.util.strings import split_arguments


class linux(Runner):
    human_name = _("Linux")
    description = _("Runs native games")
    platforms = [_("Linux")]
    entry_point_option = "exe"

    game_options = [
        {
            "option": "exe",
            "type": "file",
            "default_path": "game_path",
            "label": _("Executable"),
            "help": _("The game's main executable file"),
        },
        {
            "option": "args",
            "type": "string",
            "label": _("Arguments"),
            "help": _("Command line arguments used when launching the game"),
        },
        {
            "option": "working_dir",
            "type": "directory",
            "label": _("Working directory"),
            "help": _(
                "The location where the game is run from.\nBy default, Lutris uses the directory of the executable."
            ),
        },
        {
            "option": "ld_preload",
            "type": "file",
            "label": _("Preload library"),
            "advanced": True,
            "help": _("A library to load before running the game's executable."),
        },
        {
            "option": "ld_library_path",
            "type": "directory",
            "label": _("Add directory to LD_LIBRARY_PATH"),
            "advanced": True,
            "help": _(
                "A directory where libraries should be searched for "
                "first, before the standard set of directories; this is "
                "useful when debugging a new library or using a "
                "nonstandard library for special purposes."
            ),
        },
    ]

    def __init__(self, config=None):
        super().__init__(config)
        self.ld_preload = None

    @property
    def game_exe(self):
        """Return the game's executable's path. The file may not exist, but
        this returns None if the exe path is not defined."""
        exe = self.game_config.get("exe")
        if not exe:
            return None
        exe = os.path.expanduser(exe)  # just in case!
        if os.path.isabs(exe):
            return exe
        if self.game_path:
            return os.path.join(self.game_path, exe)
        return system.find_executable(exe)

    def resolve_game_path(self):
        return super().resolve_game_path() or os.path.dirname(self.game_exe or "")

    def get_relative_exe(self, exe_path, working_dir):
        """Return a relative path if a working dir is provided
        Some games such as Unreal Gold fail to run if given the absolute path
        """
        if exe_path and working_dir:
            relative = os.path.relpath(exe_path, start=working_dir)
            if not relative.startswith("../"):
                # We can't use the working dir implicitly to start a command
                # so we make it explicit with "./"
                if not os.path.isabs(relative):
                    relative = "./" + relative
                return relative
        return exe_path

    @property
    def working_dir(self):
        """Return the working directory to use when running the game."""
        option = self.game_config.get("working_dir")
        if option:
            return os.path.expanduser(option)
        if self.game_exe:
            return os.path.dirname(self.game_exe)
        return super().working_dir

    @property
    def nvidia_shader_cache_path(self):
        """Linux programs should get individual shader caches if possible."""
        return self.game_path or self.shader_cache_dir

    def is_installed(self, flatpak_allowed: bool = True) -> bool:
        """Well of course Linux is installed, you're using Linux right ?"""
        return True

    def can_uninstall(self):
        return False

    def uninstall(self, uninstall_callback: Callable[[], None]) -> None:
        raise RuntimeError("Linux shouldn't be installed.")

    def get_launch_config_command(self, gameplay_info, launch_config):
        # The linux runner has no command (by default) beyond the 'exe' itself;
        # so the command in gameplay_info is discarded.
        if "command" in launch_config:
            command = split_arguments(launch_config["command"])
        else:
            command = []

        working_dir = os.path.expanduser(launch_config.get("working_dir") or self.working_dir)

        if "exe" in launch_config:
            config_exe = os.path.expanduser(launch_config["exe"] or "")
            command.append(self.get_relative_exe(config_exe, working_dir))
        elif len(command) == 0:
            raise GameConfigError(_("The runner could not find a command or exe to use for this configuration."))

        if launch_config.get("args"):
            command += split_arguments(launch_config["args"])

        return command

    def get_command(self):
        # There's no command for a Linux game; the game executable
        # is the first thing in the game's command line, not any runner thing.
        return []

    def play(self):
        """Run native game."""
        launch_info = {}

        exe = self.game_exe
        if not exe or not system.path_exists(exe):
            raise MissingGameExecutableError(filename=exe)

        # Quit if the file is not executable
        mode = os.stat(exe).st_mode
        if not mode & stat.S_IXUSR:
            raise GameConfigError(_("The file %s is not executable") % exe)

        ld_preload = self.game_config.get("ld_preload")
        if ld_preload:
            launch_info["ld_preload"] = ld_preload

        ld_library_path = self.game_config.get("ld_library_path")
        if ld_library_path:
            launch_info["ld_library_path"] = os.path.expanduser(ld_library_path)

        command = [self.get_relative_exe(exe, self.working_dir)]

        args = self.game_config.get("args") or ""
        for arg in split_arguments(args):
            command.append(arg)
        launch_info["command"] = command
        return launch_info



"""Functions to interact with the Lutris REST API"""

import json
import os
import re
import socket
import time
import urllib.error
import urllib.parse
import urllib.request
from collections import OrderedDict
from datetime import datetime
from gettext import gettext as _
from typing import Any, Dict, Optional, Tuple

import requests

from lutris import settings
from lutris.gui.widgets import NotificationSource
from lutris.util import http, system
from lutris.util.graphics.gpu import get_gpus_info
from lutris.util.http import HTTPError, Request
from lutris.util.linux import LINUX_SYSTEM
from lutris.util.log import logger
from lutris.util.strings import time_ago

API_KEY_FILE_PATH = os.path.join(settings.CACHE_DIR, "auth-token")
USER_INFO_FILE_PATH = os.path.join(settings.CACHE_DIR, "user.json")


def get_time_from_api_date(date_string):
    """Convert a date string originating from the API and convert it to a datetime object"""
    return time.strptime(date_string[: date_string.find(".")], "%Y-%m-%dT%H:%M:%S")


def get_runtime_versions_date() -> float:
    return os.path.getmtime(settings.RUNTIME_VERSIONS_PATH)


def get_runtime_versions_date_time_ago() -> str:
    try:
        return time_ago(get_runtime_versions_date())
    except FileNotFoundError:
        return _("never")


def check_stale_runtime_versions() -> bool:
    """Check if the runtime needs to be updated"""
    try:
        modified_at = get_runtime_versions_date()
        should_update_at = modified_at + 6 * 60 * 60
        logger.debug(
            "Modified at %s, will update after %s",
            datetime.fromtimestamp(modified_at).strftime("%c"),
            datetime.fromtimestamp(should_update_at).strftime("%c"),
        )
        return should_update_at < time.time()
    except FileNotFoundError:
        return True


def download_runtime_versions() -> Dict[str, Any]:
    """Queries runtime + runners + current client versions and stores the result
    in a file; the mdate of this file is used to decide when it is stale and should
    be replaced."""
    gpus_info = get_gpus_info()
    pci_ids = [" ".join([gpu["PCI_ID"], gpu["PCI_SUBSYS_ID"]]) for gpu in gpus_info.values()]

    url = settings.SITE_URL + "/api/runtimes/versions?pci_ids=" + ",".join(pci_ids)
    response = http.Request(url, headers={"Content-Type": "application/json"})
    try:
        response.get()
    except http.HTTPError as ex:
        logger.error("Unable to get runtimes from API: %s", ex)
        return {}
    with open(settings.RUNTIME_VERSIONS_PATH, mode="w", encoding="utf-8") as runtime_file:
        json.dump(response.json, runtime_file, indent=2)
    return response.json


def get_runtime_versions() -> Dict[str, Any]:
    """Load runtime versions from the json file that is created at startup
    if it is missing or stale."""
    if not system.path_exists(settings.RUNTIME_VERSIONS_PATH):
        return {}
    with open(settings.RUNTIME_VERSIONS_PATH, mode="r", encoding="utf-8") as runtime_file:
        return json.load(runtime_file)


def read_api_key():
    """Read the API token from disk"""
    if not system.path_exists(API_KEY_FILE_PATH):
        return None
    with open(API_KEY_FILE_PATH, "r", encoding="utf-8") as token_file:
        api_string = token_file.read()
    try:
        username, token = api_string.split(":")
    except ValueError:
        logger.error("Unable to read Lutris token in %s", API_KEY_FILE_PATH)
        return None
    return {"token": token, "username": username}


LUTRIS_ACCOUNT_CONNECTED = NotificationSource()
LUTRIS_ACCOUNT_DISCONNECTED = NotificationSource()


def connect(username, password):
    """Connect to the Lutris API"""
    login_url = settings.SITE_URL + "/api/accounts/token"
    credentials = {"username": username, "password": password}
    try:
        response = requests.post(url=login_url, data=credentials, timeout=10)
        response.raise_for_status()
        json_dict = response.json()
        if "token" in json_dict:
            token = json_dict["token"]
            with open(API_KEY_FILE_PATH, "w", encoding="utf-8") as token_file:
                token_file.write("%s:%s" % (username, token))
            account_info = fetch_user_info()

            if not account_info:
                logger.warning("Unable to fetch user info")
            else:
                with open(USER_INFO_FILE_PATH, "w", encoding="utf-8") as token_file:
                    json.dump(account_info, token_file, indent=2)
            LUTRIS_ACCOUNT_CONNECTED.fire()
            return token
    except (
        requests.RequestException,
        requests.ConnectionError,
        requests.HTTPError,
        requests.TooManyRedirects,
        requests.Timeout,
    ) as ex:
        logger.error("Unable to connect to server (%s): %s", login_url, ex)
        return False


def disconnect():
    """Removes the API token, disconnecting the user"""
    for file_path in [API_KEY_FILE_PATH, USER_INFO_FILE_PATH]:
        if system.path_exists(file_path):
            os.remove(file_path)
    LUTRIS_ACCOUNT_DISCONNECTED.fire()


def fetch_user_info():
    """Retrieves the user info to cache it locally"""
    credentials = read_api_key()
    if not credentials:
        return
    url = settings.SITE_URL + "/api/users/me"
    request = http.Request(url, headers={"Authorization": "Token " + credentials["token"]})
    response = request.get()
    return response.json


def read_user_info():
    if not os.path.exists(USER_INFO_FILE_PATH):
        return {}
    with open(USER_INFO_FILE_PATH, encoding="utf-8") as user_info_file:
        user_info = json.load(user_info_file)
    return user_info


def get_runners(runner_name):
    """Return the available runners for a given runner name"""
    logger.debug("Retrieving runners")
    api_url = settings.SITE_URL + "/api/runners/" + runner_name
    host = settings.SITE_URL.split("//")[1]

    answers = socket.getaddrinfo(host, 443)
    (_family, _type, _proto, _canonname, _sockaddr) = answers[0]
    headers = OrderedDict({"Host": host})
    session = requests.Session()
    session.headers = headers
    response = session.get(api_url, headers=headers)
    return response.json()


def download_runner_versions(runner_name: str) -> list:
    try:
        request = Request("{}/api/runners/{}".format(settings.SITE_URL, runner_name))
        runner_info = request.get().json
        if not runner_info:
            logger.error("Failed to get runner information")
    except HTTPError as ex:
        logger.error("Unable to get runner information: %s", ex)
        runner_info = None
    if not runner_info:
        return []
    return runner_info.get("versions") or []


def format_runner_version(version_info: Dict[str, str]) -> str:
    version = version_info.get("version") or ""
    arch = version_info.get("architecture")
    return format_version_architecture(version, arch)


def normalize_version_architecture(version_name: str) -> str:
    """Ensures the version name has an architecture on the end ofit;
    adds the system's architecture if it is missing."""
    base_version, arch = parse_version_architecture(version_name)
    return format_version_architecture(base_version, arch) if base_version else ""


def format_version_architecture(base_version: str, arch: Optional[str] = None) -> str:
    """Assembles a version with architecture from the version and architecture."""
    if not base_version:
        return ""

    # A gross hack, since runner versions could be used with non-Wine runners,
    # but it so happens we don't. 'GE-Proton' versions arbitrarily do not have
    # an architecture on them - they are always 64-bit.
    if base_version.startswith("GE-Proton"):
        return base_version

    if arch:
        return "{}-{}".format(base_version, arch)

    return base_version


def parse_version_architecture(version_name: str) -> Tuple[str, Optional[str]]:
    """Splits a version that ends with an architecture into the plain version and
    architecture, as a tuple. If the version has no architecture, this provides
    the system's architecture instead."""
    if version_name:
        if version_name.endswith("-i386") or version_name.endswith("-x86_64"):
            version, arch = version_name.rsplit("-", 1)
            return version, arch
    return version_name, LINUX_SYSTEM.arch


def get_runner_version_from_cache(runner_name: str, version: Optional[str]) -> Optional[Dict[str, str]]:
    # Prefer to provide the info from our local cache if we can; if this can't find
    # an unambiguous result, we'll fall back on the API which should know what the default is.
    version, _arch = parse_version_architecture(version or "")
    runtime_versions = get_runtime_versions()
    if runtime_versions:
        try:
            runner_versions = runtime_versions["runners"][runner_name]
            runner_versions = [r for r in runner_versions if r["architecture"] in (LINUX_SYSTEM.arch, "all")]
            if version:
                runner_versions = [r for r in runner_versions if r["version"] == version]
            if len(runner_versions) == 1:
                return runner_versions[0]
        except KeyError:
            pass
    return None


def iter_get_from_api_candidates(versions: list, version: Optional[str], arch: Optional[str]):
    """A generator yielding possible version infos, or None for those that are available;
    we pick the first non-None value yielded."""

    def select_info(predicate=None, accept_ambiguous=False):
        candidates = [v for v in versions if predicate(v)]

        if candidates and (accept_ambiguous or len(candidates) == 1):
            return candidates[0]

        return None

    if version:
        yield select_info(lambda v: v["architecture"] == arch and v["version"] == version)
    else:
        yield select_info(lambda v: v["architecture"] == arch and v["default"])

    # Try various fallbacks to get some version - we prefer the default version
    # or a version with the right architecture.
    yield select_info(lambda v: v["architecture"] == arch and v["default"])
    yield select_info(lambda v: v["architecture"] == arch)

    # 64-bit system can use 32-bit version, if it's the default.
    if LINUX_SYSTEM.is_64_bit:
        yield select_info(lambda v: v["default"])

    yield select_info(lambda v: v["architecture"] == arch, accept_ambiguous=True)


def get_runner_version_from_api(runner_name: str, version: Optional[str]) -> Optional[Dict[str, str]]:
    version, arch = parse_version_architecture(version or "")
    versions = download_runner_versions(runner_name)
    for candidate in iter_get_from_api_candidates(versions, version, arch):
        if candidate:
            if not version:
                return candidate
            if version == candidate.get("version") and arch == candidate.get("architecture"):
                return candidate
    return None


def get_default_runner_version_info(runner_name: str, version: Optional[str] = None) -> Optional[Dict[str, str]]:
    """Get the appropriate version for a runner

    Params:
        version: Optional version to lookup, will return this one if found

    Returns:
        Dict containing version, architecture and url for the runner, an empty dict
        if the data can't be retrieved. If a pseudo-version is accepted, may be
        a dict containing only the version itself.
    """
    return get_runner_version_from_cache(runner_name, version) or get_runner_version_from_api(runner_name, version)


def get_http_post_response(url, payload):
    response = http.Request(url, headers={"Content-Type": "application/json"})
    try:
        response.post(data=payload)
    except http.HTTPError as ex:
        logger.error("Unable to get games from API: %s", ex)
        return None
    if response.status_code != 200:
        logger.error("API call failed: %s", response.status_code)
        return None
    return response.json


def get_game_api_page(game_slugs, page=1):
    """Read a single page of games from the API and return the response

    Args:
        game_ids (list): list of game slugs
        page (str): Page of results to get
    """
    url = settings.SITE_URL + "/api/games"
    if int(page) > 1:
        url += "?page={}".format(page)
    if not game_slugs:
        return []
    payload = json.dumps({"games": game_slugs, "page": page}).encode("utf-8")
    return get_http_post_response(url, payload)


def get_game_service_api_page(service, appids, page=1):
    """Get matching Lutris games from a list of appids from a given service"""
    url = settings.SITE_URL + "/api/games/service/%s" % service
    if int(page) > 1:
        url += "?page={}".format(page)
    if not appids:
        return []
    payload = json.dumps({"appids": appids}).encode("utf-8")
    return get_http_post_response(url, payload)


def get_api_games(game_slugs=None, page=1, service=None):
    """Return all games from the Lutris API matching the given game slugs"""
    if service:
        response_data = get_game_service_api_page(service, game_slugs)
    else:
        response_data = get_game_api_page(game_slugs)

    if not response_data:
        return []
    results = response_data.get("results", [])
    while response_data.get("next"):
        page_match = re.search(r"page=(\d+)", response_data["next"])
        if page_match:
            next_page = page_match.group(1)
        else:
            logger.error("No page found in %s", response_data["next"])
            break
        if service:
            response_data = get_game_service_api_page(service, game_slugs, page=next_page)
        else:
            response_data = get_game_api_page(game_slugs, page=next_page)
        if not response_data:
            logger.warning("Unable to get response for page %s", next_page)
            break
        results += response_data.get("results")
    return results


def get_game_installers(game_slug, revision=None):
    """Get installers for a single game"""
    if not game_slug:
        raise ValueError("No game_slug provided. Can't query an installer")
    if revision:
        installer_url = settings.INSTALLER_REVISION_URL % (game_slug, revision)
    else:
        installer_url = settings.INSTALLER_URL % game_slug

    logger.debug("Fetching installer %s", installer_url)
    request = http.Request(installer_url)
    request.get()
    response = request.json
    if response is None:
        raise RuntimeError("Couldn't get installer at %s" % installer_url)

    # Revision requests return a single installer
    if revision:
        installers = [response]
    else:
        installers = response["results"]
    return [normalize_installer(i) for i in installers]


def get_game_details(slug: str) -> dict:
    url = settings.SITE_URL + "/api/games/%s" % slug
    request = http.Request(url)
    try:
        response = request.get()
    except http.HTTPError as ex:
        logger.debug("Unable to load %s: %s", slug, ex)
        return {}
    return response.json


def normalize_installer(installer: dict, **additionnal) -> dict:
    """Adjusts an installer dict so it is in the correct form, with values
    of the expected types. ``additionnal`` kwargs will be added to the result as is."""

    def must_be_str(key):
        if key in installer:
            installer[key] = str(installer[key])

    must_be_str("name")
    must_be_str("version")
    must_be_str("os")
    must_be_str("slug")
    must_be_str("game_slug")
    must_be_str("dlcid")
    must_be_str("runner")
    installer.update(additionnal)
    return installer


def search_games(query) -> dict:
    if not query:
        return {}
    query = query.lower().strip()[:255]
    url = "/api/games?%s" % urllib.parse.urlencode({"search": query, "with-installers": True})
    response = http.Request(settings.SITE_URL + url, headers={"Content-Type": "application/json"})
    try:
        response.get()
    except http.HTTPError as ex:
        logger.error("Unable to get games from API: %s", ex)
        return {}
    return response.json


def parse_installer_url(url):
    """
    Parses `lutris:` urls, extracting any info necessary to install or run a game.
    """
    action = None
    launch_config_name = None
    try:
        parsed_url = urllib.parse.urlparse(url, scheme="lutris")
    except Exception as ex:
        logger.warning("Unable to parse url %s", url)
        raise ValueError("Invalid lutris url %s" % url) from ex
    if parsed_url.scheme != "lutris":
        raise ValueError("Invalid lutris url %s" % url)
    url_path = parsed_url.path
    if not url_path:
        raise ValueError("Invalid lutris url %s" % url)
    # urlparse can't parse if the path only contain numbers
    # workaround to remove the scheme manually:
    if url_path.startswith("lutris:"):
        url_path = url_path[7:]

    url_parts = [urllib.parse.unquote(part) for part in url_path.split("/")]
    if len(url_parts) == 3:
        action = url_parts[0]
        game_slug = url_parts[1]
        launch_config_name = url_parts[2]
    elif len(url_parts) == 2:
        action = url_parts[0]
        game_slug = url_parts[1]
    elif len(url_parts) == 1:
        game_slug = url_parts[0]
    else:
        raise ValueError("Invalid lutris url %s" % url)

    # To link to service games, format a slug like <service>:<appid>
    if ":" in game_slug:
        service, appid = game_slug.split(":", maxsplit=1)
    else:
        service, appid = "", ""

    revision = None
    if parsed_url.query:
        query = dict(urllib.parse.parse_qsl(parsed_url.query))
        revision = query.get("revision")
    return {
        "game_slug": game_slug,
        "revision": revision,
        "action": action,
        "service": service,
        "appid": appid,
        "launch_config_name": launch_config_name,
    }


def format_installer_url(installer_info):
    """
    Generates 'lutris:' urls, given the same dictionary that
    parse_intaller_url returns.
    """

    game_slug = installer_info.get("game_slug")
    revision = installer_info.get("revision")
    action = installer_info.get("action")
    service = installer_info.get("service")
    appid = installer_info.get("appid")
    launch_config_name = installer_info.get("launch_config_name")
    parts = []

    if action:
        parts.append(action)
    elif not launch_config_name:
        raise ValueError("A 'lutris:' URL can contain a launch configuration name only if it has an action.")

    if game_slug:
        parts.append(game_slug)
    else:
        parts.append(service + ":" + appid)

    if launch_config_name:
        parts.append(launch_config_name)

    parts = [urllib.parse.quote(str(part), safe="") for part in parts]
    path = "/".join(parts)

    if revision:
        query = urllib.parse.urlencode({"revision": str(revision)})
    else:
        query = ""

    url = urllib.parse.urlunparse(("lutris", "", path, "", query, None))
    return url



"""Move hidden games from settings to database"""

from lutris import settings
from lutris.game import Game


def get_hidden_ids():
    """Return a list of game IDs to be excluded from the library view"""
    # Load the ignore string and filter out empty strings to prevent issues
    ignores_raw = settings.read_setting("library_ignores").split(",")
    return [ignore.strip() for ignore in ignores_raw if ignore]


def migrate():
    """Run migration"""
    try:
        game_ids = get_hidden_ids()
    except:
        print("Failed to read hidden game IDs")
        return []
    for game_id in game_ids:
        game = Game(game_id)
        game.mark_as_hidden(True)
    settings.write_setting("library_ignores", "", section="lutris")



"""Read and write VDF files"""

# Lutris Modules
from lutris.util.log import logger


def vdf_parse(steam_config_file, config):
    """Parse a Steam config file and return the contents as a dict."""
    line = " "
    while line:
        try:
            line = steam_config_file.readline()

        except UnicodeDecodeError:
            logger.error(
                "Error while reading Steam VDF file %s. Returning %s",
                steam_config_file,
                config,
            )

            return config
        if not line or line.strip() == "}":
            return config
        while not line.strip().endswith('"'):
            nextline = steam_config_file.readline()
            if not nextline:
                break
            line = line[:-1] + nextline
        # Handle escaped quotes in values by temporarily replacing them
        # This allows split('"') to work correctly
        line = line.replace('\\"', "\x00")  # Use null as placeholder
        line_elements = line.strip().split('"')
        # Restore escaped quotes in the parsed values
        line_elements = [elem.replace("\x00", '"') for elem in line_elements]
        if len(line_elements) == 3:
            key = line_elements[1]
            steam_config_file.readline()  # skip '{'
            config[key] = vdf_parse(steam_config_file, {})
        else:
            try:
                config[line_elements[1]] = line_elements[3]
            except IndexError:
                logger.error("Malformed config file: %s", line)
    return config


def to_vdf(dict_data, level=0):
    """Convert a dictionnary to Steam config file format"""
    vdf_data = ""
    for key in dict_data:
        value = dict_data[key]
        if isinstance(value, dict):
            vdf_data += '%s"%s"\n' % ("\t" * level, key)
            vdf_data += "%s{\n" % ("\t" * level)
            vdf_data += to_vdf(value, level + 1)
            vdf_data += "%s}\n" % ("\t" * level)
        else:
            vdf_data += '%s"%s"\t\t"%s"\n' % ("\t" * level, key, value)
    return vdf_data


def vdf_write(vdf_path, config):
    """Write a Steam configuration to a vdf file"""
    vdf_data = to_vdf(config)
    with open(vdf_path, "w", encoding="utf-8") as vdf_file:
        vdf_file.write(vdf_data)



import filecmp
import os
from gettext import gettext as _
from shutil import copyfile

from lutris.exceptions import MissingGameExecutableError
from lutris.runners.runner import Runner
from lutris.util import system
from lutris.util.log import logger


class ryujinx(Runner):
    human_name = _("Ryujinx")
    platforms = [_("Nintendo Switch")]
    description = _("Nintendo Switch emulator")
    runnable_alone = True
    runner_executable = "ryujinx/publish/Ryujinx"
    flatpak_id = "io.github.ryubing.Ryujinx"
    download_url = "https://lutris.nyc3.digitaloceanspaces.com/runners/ryujinx/ryujinx-1.0.7074-linux_x64.tar.gz"

    game_options = [
        {
            "option": "main_file",
            "type": "file",
            "label": _("NSP file"),
            "help": _("The game data, commonly called a ROM image."),
        },
        {
            "option": "application_id",
            "type": "string",
            "label": _("Application ID"),
            "help": _("Select which app to launch for a multi-game XCI image"),
        },
    ]
    runner_options = [
        {
            "option": "prod_keys",
            "label": _("Encryption keys"),
            "type": "file",
            "help": _("File containing the encryption keys."),
        },
        {
            "option": "title_keys",
            "label": _("Title keys"),
            "type": "file",
            "help": _("File containing the title keys."),
        },
    ]

    @property
    def ryujinx_data_dir(self):
        """Return dir where Ryujinx files lie."""
        candidates = ("~/.local/share/ryujinx",)
        for candidate in candidates:
            path = system.fix_path_case(os.path.join(os.path.expanduser(candidate), "nand"))
            if system.path_exists(path):
                return path[: -len("nand")]

    def play(self):
        """Run the game."""
        arguments = self.get_command()
        rom = self.game_config.get("main_file") or ""
        if not system.path_exists(rom):
            raise MissingGameExecutableError(filename=rom)

        if app_id := self.game_config.get("application_id"):
            arguments.append("--application-id")
            arguments.append(app_id)
        arguments.append(rom)
        return {"command": arguments}

    def _update_key(self, key_type):
        """Update a keys file if set"""
        ryujinx_data_dir = self.ryujinx_data_dir
        if not ryujinx_data_dir:
            logger.error("Ryujinx data dir not set")
            return
        if key_type == "prod_keys":
            key_loc = os.path.join(ryujinx_data_dir, "keys/prod.keys")
        elif key_type == "title_keys":
            key_loc = os.path.join(ryujinx_data_dir, "keys/title.keys")
        else:
            logger.error("Invalid keys type %s!", key_type)
            return

        key = self.runner_config.get(key_type)
        if not key:
            logger.debug("No %s file was set.", key_type)
            return
        if not system.path_exists(key):
            logger.warning("Keys file %s does not exist!", key)
            return

        keys_dir = os.path.dirname(key_loc)
        if not os.path.exists(keys_dir):
            os.makedirs(keys_dir)
        elif os.path.isfile(key_loc) and filecmp.cmp(key, key_loc):
            # If the files are identical, don't do anything
            return
        copyfile(key, key_loc)

    def prelaunch(self):
        for key in ["prod_keys", "title_keys"]:
            self._update_key(key_type=key)



import os
from gettext import gettext as _

from gi.repository import Gtk

from lutris.cache import get_custom_cache_path, save_custom_cache_path, validate_custom_cache_path
from lutris.config import LutrisConfig
from lutris.gui.config.base_config_box import BaseConfigBox
from lutris.gui.config.widget_generator import WidgetWarningMessageBox
from lutris.gui.widgets.common import FileChooserEntry, Label
from lutris.runners.runner import Runner
from lutris.util.jobs import AsyncCall
from lutris.util.retroarch.firmware import scan_firmware_directory
from lutris.util.strings import human_size


class StorageBox(BaseConfigBox):
    def populate(self):
        self.error_boxes = {
            "bios_path": StoragePathMessageBox(),
            "pga_cache_path": StoragePathMessageBox(),
        }

        self.add(self.get_section_label(_("Paths")))
        path_widgets = self.get_path_widgets()
        self.pack_start(self._get_framed_options_list_box(path_widgets), False, False, 0)
        self.update_pga_cache_path_warning()

    def get_path_widgets(self):
        widgets = []
        base_runner = Runner()
        bios_path = base_runner.config.system_config.get("bios_path")

        path_settings = [
            {
                "name": _("Game library"),
                "setting": "game_path",
                "default": os.path.expanduser("~/Games"),
                "value": base_runner.default_path,
                "help": _("The default folder where you install your games."),
            },
            {
                "name": _("Installer cache"),
                "setting": "pga_cache_path",
                "default": "",
                "value": get_custom_cache_path() or "",
                "help": _(
                    "If provided, files downloaded during game installs will be kept there\n"
                    "\nOtherwise, all downloaded files are discarded."
                ),
            },
            {
                "name": _("Emulator BIOS files location"),
                "setting": "bios_path",
                "default": "",
                "value": bios_path if bios_path else "",
                "help": _("The folder Lutris will search in for emulator BIOS files if needed"),
            },
        ]
        for path_setting in path_settings:
            widgets.append(self.get_directory_chooser(path_setting))
        return widgets

    def get_directory_chooser(self, path_setting):
        label = Label()
        label.set_markup("<b>%s</b>" % path_setting["name"])
        wrapper = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=4, visible=True)
        wrapper.set_margin_top(16)

        default_path = path_setting["default"]
        directory_chooser = FileChooserEntry(
            title=_("Select folder"),
            action=Gtk.FileChooserAction.SELECT_FOLDER,
            warn_if_non_writable_parent=True,
            text=path_setting["value"],
            default_path=default_path,
        )
        directory_chooser.connect("changed", self.on_file_chooser_changed, path_setting)
        wrapper.pack_start(label, False, False, 0)
        wrapper.pack_start(directory_chooser, True, True, 0)
        if path_setting["help"]:
            help_wrapper = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=4, visible=True)
            help_wrapper.add(wrapper)
            help_label = Label()
            help_label.set_markup("<i>%s</i>" % path_setting["help"])
            help_wrapper.add(help_label)
            wrapper = help_wrapper

        if path_setting["setting"] in self.error_boxes:
            warning = self.error_boxes[path_setting["setting"]]
            wrapper.add(warning)

        wrapper.set_margin_end(16)
        wrapper.set_margin_left(16)
        wrapper.set_margin_bottom(16)

        return wrapper

    def is_bios_path_invalid(self, bios_path):
        if not bios_path:
            return bios_path, ""  # it's fine to not have a BIOS path
        MAX_BIOS_FOLDER_SIZE = 5e9
        MAX_BIOS_FILES_IN_FOLDER = 5000
        MAX_BIOS_FOLDER_DEPTH = 3

        bios_path_size = 0
        bios_path_file_count = 0
        bios_path_depth = 0

        for path, _dir_names, file_names in os.walk(bios_path):
            for file_name in file_names:
                file_path = f"{path}/{file_name}"
                if os.access(file_path, os.R_OK):
                    bios_path_size += os.path.getsize(file_path)
                    bios_path_file_count += 1
                    bios_path_depth = path[len(bios_path) :].count(os.sep)

        if bios_path_size > MAX_BIOS_FOLDER_SIZE:
            return bios_path, _("Folder is too large (%s)") % human_size(bios_path_size)
        if bios_path_file_count > MAX_BIOS_FILES_IN_FOLDER:
            return bios_path, _("Too many files in folder")
        if bios_path_depth > MAX_BIOS_FOLDER_DEPTH:
            return bios_path, _("Folder is too deep")
        return bios_path, ""

    def bios_path_validated_cb(self, result, error):
        error_box = self.error_boxes["bios_path"]

        if error:
            error_box.show_markup(None)
            return

        bios_path, error_message = result

        error_box.show_markup(error_message)

        if not error_message:
            lutris_config = LutrisConfig()
            lutris_config.raw_system_config["bios_path"] = bios_path
            lutris_config.save()
            AsyncCall(scan_firmware_directory, None, bios_path)

    def on_file_chooser_changed(self, entry, setting):
        folder_path = entry.get_text()
        if setting["setting"] == "pga_cache_path":
            save_custom_cache_path(folder_path)
            self.update_pga_cache_path_warning()
        elif setting["setting"] == "game_path":
            lutris_config = LutrisConfig()
            lutris_config.raw_system_config["game_path"] = folder_path
            lutris_config.save()
        elif setting["setting"] == "bios_path":
            AsyncCall(self.is_bios_path_invalid, self.bios_path_validated_cb, folder_path)

    def update_pga_cache_path_warning(self):
        cache_path = get_custom_cache_path()
        if cache_path:
            valid, markup = validate_custom_cache_path(cache_path)
        else:
            valid, markup = True, None

        error_box = self.error_boxes["pga_cache_path"]
        error_box.show_markup(markup if markup and not valid else None)


class StoragePathMessageBox(WidgetWarningMessageBox):
    def __init__(self, icon_name="dialog-error"):
        super().__init__(icon_name=icon_name)



import logging
import os.path
from gettext import gettext as _

from lutris.config import LutrisConfig
from lutris.exceptions import MissingBiosError, MissingGameExecutableError
from lutris.runners.runner import Runner
from lutris.util import display, extract, system


def get_resolutions():
    try:
        screen_resolutions = [(resolution, resolution) for resolution in display.DISPLAY_MANAGER.get_resolutions()]
    except OSError:
        screen_resolutions = []
    screen_resolutions.insert(0, (_("Desktop resolution"), "desktop"))
    return screen_resolutions


class atari800(Runner):
    human_name = _("Atari800")
    platforms = [_("Atari 8bit computers")]  # FIXME try to determine the actual computer used
    runner_executable = "atari800/bin/atari800"
    bios_url = "https://netactuate.dl.sourceforge.net/project/atari800/ROM/Original%20XL%20ROM/xf25.zip?viasf=1"
    description = _("Atari 400, 800 and XL emulator")
    bios_checksums = {
        "xlxe_rom": "06daac977823773a3eea3422fd26a703",
        "basic_rom": "0bac0c6a50104045d902df4503a4c30b",
        "osa_rom": "",
        "osb_rom": "a3e8d617c95d08031fe1b20d541434b2",
        "5200_rom": "",
    }
    game_options = [
        {
            "option": "main_file",
            "type": "file",
            "label": _("ROM file"),
            "help": _(
                "The game data, commonly called a ROM image. \n"
                "Supported formats: ATR, XFD, DCM, ATR.GZ, XFD.GZ "
                "and PRO."
            ),
        }
    ]

    runner_options = [
        {
            "option": "bios_path",
            "type": "directory",
            "label": _("BIOS location"),
            "help": _(
                "A folder containing the Atari 800 BIOS files.\n"
                "They are provided by Lutris so you shouldn't have to "
                "change this."
            ),
        },
        {
            "option": "machine",
            "type": "choice",
            "choices": [
                (_("Emulate Atari 800"), "atari"),
                (_("Emulate Atari 800 XL"), "xl"),
                (_("Emulate Atari 320 XE (Compy Shop)"), "320xe"),
                (_("Emulate Atari 320 XE (Rambo)"), "rambo"),
                (_("Emulate Atari 5200"), "5200"),
            ],
            "default": "atari",
            "label": _("Machine"),
        },
        {
            "option": "fullscreen",
            "type": "bool",
            "default": False,
            "section": _("Graphics"),
            "label": _("Fullscreen"),
        },
        {
            "option": "resolution",
            "type": "choice",
            "choices": get_resolutions(),
            "default": "desktop",
            "section": _("Graphics"),
            "label": _("Fullscreen resolution"),
        },
    ]

    def install(self, install_ui_delegate, version=None, callback=None):
        def on_runner_installed(*_args):
            config_path = system.create_folder("~/.atari800")
            bios_archive = os.path.join(config_path, "atari800-bioses.zip")
            install_ui_delegate.download_install_file(self.bios_url, bios_archive)
            if not system.path_exists(bios_archive):
                raise RuntimeError(_("Could not download Atari 800 BIOS archive"))
            extract.extract_archive(bios_archive, config_path)
            os.remove(bios_archive)
            config = LutrisConfig(runner_slug="atari800")
            config.raw_runner_config.update({"bios_path": config_path})
            config.save()
            if callback:
                callback()

        super().install(install_ui_delegate, version, on_runner_installed)

    def find_good_bioses(self, bios_path):
        """Check for correct bios files"""
        good_bios = {}
        for filename in os.listdir(bios_path):
            real_hash = system.get_md5_hash(os.path.join(bios_path, filename))
            for bios_file, checksum in self.bios_checksums.items():
                if real_hash == checksum:
                    logging.debug("%s Checksum : OK", filename)
                    good_bios[bios_file] = filename
        return good_bios

    def play(self):
        arguments = self.get_command()
        if self.runner_config.get("fullscreen"):
            arguments.append("-fullscreen")
        else:
            arguments.append("-windowed")

        resolution = self.runner_config.get("resolution")
        if resolution:
            if resolution == "desktop":
                width, height = display.DISPLAY_MANAGER.get_current_resolution()
            else:
                width, height = resolution.split("x")
            arguments += ["-fs-width", "%s" % width, "-fs-height", "%s" % height]

        if self.runner_config.get("machine"):
            arguments.append("-%s" % self.runner_config["machine"])

        bios_path = self.runner_config.get("bios_path")
        if not system.path_exists(bios_path):
            raise MissingBiosError()
        good_bios = self.find_good_bioses(bios_path)
        for bios, filename in good_bios.items():
            arguments.append("-%s" % bios)
            arguments.append(os.path.join(bios_path, filename))

        rom = self.game_config.get("main_file") or ""
        if not system.path_exists(rom):
            raise MissingGameExecutableError(filename=rom)
        arguments.append(rom)

        return {"command": arguments}



import sqlite3
import threading
from typing import Any

# Prevent multiple access to the database (SQLite limitation)
DB_LOCK = threading.RLock()


class db_cursor(object):
    def __init__(self, db_path):
        self.db_path = db_path
        self.db_conn = None

    def __enter__(self):
        self.db_conn = sqlite3.connect(self.db_path)
        cursor = self.db_conn.cursor()
        return cursor

    def __exit__(self, _type, value, traceback):
        self.db_conn.commit()
        self.db_conn.close()


def cursor_execute(cursor, query, params=None):
    """Execute a SQL query, run it in a lock block"""
    params = params or ()
    lock = DB_LOCK.acquire(timeout=5)  # pylint: disable=consider-using-with
    if not lock:
        raise RuntimeError(f"Database is busy. Not executing {query}")

    try:
        return cursor.execute(query, params)
    finally:
        DB_LOCK.release()


def db_insert(db_path, table, fields):
    columns = ", ".join(list(fields.keys()))
    placeholders = ("?, " * len(fields))[:-2]
    field_values = tuple(fields.values())
    with db_cursor(db_path) as cursor:
        cursor_execute(
            cursor,
            "insert into {0}({1}) values ({2})".format(table, columns, placeholders),
            field_values,
        )
        inserted_id = cursor.lastrowid
    return inserted_id


def db_update(db_path, table, updated_fields, conditions):
    """Update `table` with the values given in the dict `values` on the
    condition given with the `row` tuple.
    """
    columns = "=?, ".join(list(updated_fields.keys())) + "=?"
    field_values = tuple(updated_fields.values())

    condition_field = " AND ".join(["%s=?" % field for field in conditions])
    condition_value = tuple(conditions.values())

    with db_cursor(db_path) as cursor:
        query = "UPDATE {0} SET {1} WHERE {2}".format(table, columns, condition_field)
        result = cursor_execute(cursor, query, field_values + condition_value)
    return result


def db_delete(db_path, table, field, value):
    with db_cursor(db_path) as cursor:
        cursor_execute(cursor, "delete from {0} where {1}=?".format(table, field), (value,))


def db_select(db_path, table, fields=None, condition=None):
    if fields:
        columns = ", ".join(fields)
    else:
        columns = "*"
    with db_cursor(db_path) as cursor:
        query = "SELECT {} FROM {}"
        if condition:
            condition_field, condition_value = condition
            if isinstance(condition_value, (list, tuple, set)):
                condition_value = tuple(condition_value)
                placeholders = ", ".join("?" * len(condition_value))
                where_condition = " where {} in (" + placeholders + ")"
            else:
                condition_value = (condition_value,)
                where_condition = " where {}=?"
            query = query + where_condition
            query = query.format(columns, table, condition_field)
            params = condition_value
        else:
            query = query.format(columns, table)
            params = ()
        cursor_execute(cursor, query, params)
        rows = cursor.fetchall()
        column_names = [column[0] for column in cursor.description]
    results = []
    for row in rows:
        row_data = {}
        for index, column in enumerate(column_names):
            row_data[column] = row[index]
        results.append(row_data)
    return results


def db_query(db_path, query, params=()):
    with db_cursor(db_path) as cursor:
        cursor_execute(cursor, query, params)
        rows = cursor.fetchall()
        column_names = [column[0] for column in cursor.description]
    results = []
    for row in rows:
        row_data = {}
        for index, column in enumerate(column_names):
            row_data[column] = row[index]
        results.append(row_data)
    return results


def add_field(db_path, tablename, field):
    query = "ALTER TABLE %s ADD COLUMN %s %s" % (
        tablename,
        field["name"],
        field["type"],
    )
    with db_cursor(db_path) as cursor:
        cursor.execute(query)


def _create_filter(field: str, value: Any, params: list, negate: bool = False) -> str:
    """Creates a filter to match a field to a vlaue, or to a list of
    values. None can be used as well, to make NULL."""
    also_null = False
    if hasattr(value, "__iter__") and not isinstance(value, str):
        values = list(value)

        if None in values:
            values.remove(None)
            also_null = True
    elif value is None:
        also_null = True
        values = []
    else:
        values = [value]

    if len(values) == 0:
        if negate:
            if also_null:
                return f"{field} IS NOT NULL"
            else:
                return "1 = 1"
        else:
            if also_null:
                return f"{field} IS NULL"
            else:
                return "1 = 0"

    if len(values) == 1:
        params.append(values[0])
        sql = f"{field} != ?" if negate else f"{field} = ?"
    else:
        sql = f"{field} NOT IN (" if negate else f"{field} IN ("
        for i, v in enumerate(values):
            params.append(v)
            if i > 0:
                sql += ", "
            sql += "?"
        sql += ")"

    if also_null:
        if negate:
            return f"({field} IS NOT NULL AND {sql})"
        else:
            return f"({field} IS NULL OR {sql})"
    else:
        return sql


def filtered_query(db_path, table, searches=None, filters=None, excludes=None, sorts=None):
    query = "select * from %s" % table
    params = []
    sql_filters = []
    for field in searches or {}:
        sql_filters.append("%s LIKE ?" % field)
        params.append("%" + searches[field] + "%")
    for field in filters or {}:
        sql_filters.append(_create_filter(field, filters[field], params))
    for field in excludes or {}:
        sql_filters.append(_create_filter(field, excludes[field], params, negate=True))
    if sql_filters:
        query += " WHERE " + " AND ".join(sql_filters)
    if sorts:
        query += " ORDER BY %s" % ", ".join(["%s %s" % (sort[0], sort[1]) for sort in sorts])
    else:
        query += " ORDER BY slug ASC"
    return db_query(db_path, query, tuple(params))



"""Valve Sniper runtime (Steam Linux Runtime 3.0) discovery utilities."""

import os
from functools import lru_cache
from typing import List, Optional

from lutris.util.log import logger

_SNIPER_SEARCH_PATHS = [
    os.path.expanduser("~/.local/share/umu/run-in-sniper"),
    os.path.expanduser("~/.local/share/Steam/steamapps/common/SteamLinuxRuntime_sniper/run-in-sniper"),
]

# Host library base directories. Inside the Sniper container, the host
# filesystem is mounted at /run/host/.
_HOST_LIB_DIRS = ["/usr/lib64", "/usr/lib", "/lib64", "/lib"]

# Sniper runtime library directories inside the container. These must come
# BEFORE host fallback paths in LD_LIBRARY_PATH so that libraries present
# in Sniper (e.g. gdk-pixbuf with its module loaders) are preferred over
# the host's versions, which would have wrong module paths inside the container.
_SNIPER_RUNTIME_LIB_DIRS = [
    "/usr/lib/x86_64-linux-gnu",
    "/lib/x86_64-linux-gnu",
    "/usr/lib/i386-linux-gnu",
    "/lib/i386-linux-gnu",
]


def get_sniper_run_command() -> Optional[str]:
    """Return the path to the run-in-sniper script, or None if not found.

    Searches in order:
      1. umu-launcher installation (~/.local/share/umu/)
      2. Steam installation (~/.local/share/Steam/)
    """
    for path in _SNIPER_SEARCH_PATHS:
        if os.path.isfile(path) and os.access(path, os.X_OK):
            logger.debug("Found Sniper runtime at %s", path)
            return path

    logger.debug("Sniper runtime not found")
    return None


@lru_cache(maxsize=1)
def get_sniper_host_lib_paths() -> List[str]:
    """Discover host library directories and return them as /run/host/ paths
    for use inside the Sniper container.

    Scans base library dirs and their immediate subdirectories for directories
    that contain shared libraries (.so files). This catches private library
    directories like /usr/lib64/pulseaudio/ that use RUNPATH and would otherwise
    resolve to the Sniper runtime's version instead of the host's.
    """
    seen = set()
    paths = []

    for host_dir in _HOST_LIB_DIRS:
        real_dir = os.path.realpath(host_dir)
        if real_dir in seen or not os.path.isdir(real_dir):
            continue
        seen.add(real_dir)

        container_dir = "/run/host" + host_dir
        paths.append(container_dir)

        try:
            entries = os.listdir(real_dir)
        except OSError:
            continue

        for entry in entries:
            subdir = os.path.join(real_dir, entry)
            if not os.path.isdir(subdir):
                continue
            try:
                if any(f.endswith(".so") or ".so." in f for f in os.listdir(subdir)):
                    paths.append("/run/host" + os.path.join(host_dir, entry))
            except OSError:
                continue

    return paths


def get_sniper_ld_library_path() -> str:
    """Build the LD_LIBRARY_PATH for use inside the Sniper container.

    The order ensures:
      1. Sniper runtime libs (preferred — modules like gdk-pixbuf loaders work correctly)
      2. Host libs via /run/host/ (fallback — for libraries missing from Sniper like libgtkmm)
    """
    return os.pathsep.join(_SNIPER_RUNTIME_LIB_DIRS + get_sniper_host_lib_paths())



"""Store object for a list of games"""

# pylint: disable=not-an-iterable
import time
from typing import Set, Union

from gi.repository import GLib, GObject, Gtk

from lutris import settings
from lutris.database import sql
from lutris.database.games import get_all_installed_game_for_service, get_games
from lutris.gui.views.store_item import StoreItem
from lutris.util.strings import gtk_safe

from . import (
    COL_ID,
    COL_INSTALLED,
    COL_INSTALLED_AT,
    COL_INSTALLED_AT_TEXT,
    COL_LASTPLAYED,
    COL_LASTPLAYED_TEXT,
    COL_MEDIA_PATHS,
    COL_NAME,
    COL_PLATFORM,
    COL_PLAYTIME,
    COL_PLAYTIME_TEXT,
    COL_RUNNER,
    COL_RUNNER_HUMAN_NAME,
    COL_SLUG,
    COL_SORTNAME,
    COL_YEAR,
)


def try_lower(value):
    try:
        out = value.lower()
    except AttributeError:
        out = value
    return out


def sort_func(model, row1, row2, sort_col):
    """Sorting function for the game store"""
    value1 = model.get_value(row1, sort_col)
    value2 = model.get_value(row2, sort_col)
    if value1 is None and value2 is None:
        value1 = value2 = 0
    elif value1 is None:
        value1 = type(value2)()
    elif value2 is None:
        value2 = type(value1)()
    value1 = try_lower(value1)
    value2 = try_lower(value2)
    diff = -1 if value1 < value2 else 0 if value1 == value2 else 1
    if diff == 0:
        value1 = try_lower(model.get_value(row1, COL_SORTNAME))
        value2 = try_lower(model.get_value(row2, COL_SORTNAME))
        try:
            diff = -1 if value1 < value2 else 0 if value1 == value2 else 1
        except TypeError:
            diff = 0
    if diff == 0:
        value1 = try_lower(model.get_value(row1, COL_RUNNER_HUMAN_NAME))
        value2 = try_lower(model.get_value(row2, COL_RUNNER_HUMAN_NAME))
    try:
        return -1 if value1 < value2 else 0 if value1 == value2 else 1
    except TypeError:
        return 0


class GameStore(GObject.Object):
    def __init__(self, service, service_media):
        super().__init__()
        self.service = service
        self.service_media = service_media
        self._installed_games = []
        self._installed_games_accessed = False
        self._icon_updates = {}

        self.store = Gtk.ListStore(
            str,
            str,
            str,
            str,
            GObject.TYPE_PYOBJECT,
            str,
            str,
            str,
            str,
            int,
            str,
            bool,
            int,
            str,
            float,
            str,
        )

    @property
    def installed_game_slugs(self):
        previous_access = self._installed_games_accessed or 0
        self._installed_games_accessed = time.time()
        if self._installed_games_accessed - previous_access > 1:
            self._installed_games = [g["slug"] for g in get_games(filters={"installed": "1"})]
        return self._installed_games

    def get_row_by_slug(self, slug):
        for model_row in self.store:
            if model_row[COL_SLUG] == slug:
                return model_row

    def get_row_by_id(self, _id):
        if not _id:
            return
        for model_row in self.store:
            try:
                if model_row[COL_ID] == str(_id):
                    return model_row
            except TypeError:
                return

    def remove_game(self, _id):
        """Remove a game from the view."""
        row = self.get_row_by_id(_id)
        if row:
            self.store.remove(row.iter)

    def update(self, db_game: dict) -> Union[Set[int], None]:
        """Update game information
        Return the indices of the row that were updated, or an empty set if no change
        was made, or None if the game could not be found.
        """
        store_item = StoreItem(db_game, self.service, self.service_media)
        row = self.get_row_by_id(store_item.id)
        if not row and "service_id" in db_game:
            row = self.get_row_by_id(db_game["service_id"])
        if not row:
            return None

        new_values = dict()

        new_values[COL_ID] = str(store_item.id)
        new_values[COL_SLUG] = store_item.slug
        new_values[COL_NAME] = store_item.name
        new_values[COL_SORTNAME] = store_item.sortname if store_item.sortname else store_item.name
        new_values[COL_MEDIA_PATHS] = store_item.get_media_paths() if settings.SHOW_MEDIA else []
        new_values[COL_YEAR] = store_item.year
        new_values[COL_RUNNER] = store_item.runner
        new_values[COL_RUNNER_HUMAN_NAME] = store_item.runner_text
        new_values[COL_PLATFORM] = store_item.platform
        new_values[COL_LASTPLAYED] = store_item.lastplayed
        new_values[COL_LASTPLAYED_TEXT] = store_item.lastplayed_text
        new_values[COL_INSTALLED] = store_item.installed
        new_values[COL_INSTALLED_AT] = store_item.installed_at
        new_values[COL_INSTALLED_AT_TEXT] = store_item.installed_at_text
        new_values[COL_PLAYTIME] = store_item.playtime
        new_values[COL_PLAYTIME_TEXT] = store_item.playtime_text

        changed_indices = set()
        for idx, value in new_values.items():
            if row[idx] != value:
                row[idx] = value
                changed_indices.add(idx)
        return changed_indices

    def add_game(self, db_game):
        """Add a game to the store"""
        store_item = StoreItem(db_game, self.service, self.service_media)
        self.add_item(store_item)

    def add_item(self, store_item):
        self.store.append(
            (
                store_item.id,
                store_item.slug,
                store_item.name,
                store_item.sortname if store_item.sortname else store_item.name,
                store_item.get_media_paths() if settings.SHOW_MEDIA else [],
                store_item.year,
                store_item.runner,
                store_item.runner_text,
                gtk_safe(store_item.platform),
                store_item.lastplayed,
                store_item.lastplayed_text,
                store_item.installed,
                store_item.installed_at,
                store_item.installed_at_text,
                store_item.playtime,
                store_item.playtime_text,
            )
        )

    def add_preloaded_games(self, db_games, service_id):
        """Add games to the store, but preload their installed-game data
        all at once, for faster database access. This should be used if all or almost all
        games are being loaded."""

        installed_db_games = {}
        if service_id and db_games:
            installed_db_games = get_all_installed_game_for_service(service_id)

        for db_game in db_games:
            if installed_db_games is not None and "appid" in db_game:
                appid = db_game["appid"]
                store_item = StoreItem(db_game, self.service, self.service_media)
                store_item.apply_installed_game_data(installed_db_games.get(appid))
                self.add_item(store_item)
            else:
                self.add_game(db_game)

    def on_game_updated(self, game):
        if self.service:
            db_games = sql.filtered_query(
                settings.DB_PATH,
                "service_games",
                filters=({"service": self.service, "appid": game.appid}),
            )
        else:
            db_games = sql.filtered_query(settings.DB_PATH, "games", filters=({"id": game.id}))

        for db_game in db_games:
            GLib.idle_add(self.update, db_game)
        return True



from gi.repository import Gtk


def update_action_widget_visibility(widgets, visible_predicate):
    """This sets the visibility on a set of widgets, like menu items. You provide a function
    that indicates if an item is visible, or None for separators that are visible based on
    their neighbors. Returns the count of visible widgets that are not separators."""
    visible_count = 0
    previous_visible_widget = None
    for w in widgets:
        visible = visible_predicate(w)

        if visible:
            visible_count = visible_count + 1

        if visible is None:
            if previous_visible_widget is None:
                visible = False
            else:
                visible = visible_predicate(previous_visible_widget) is not None

        w.set_visible(visible)
        if visible:
            previous_visible_widget = w

    if previous_visible_widget and visible_predicate(previous_visible_widget) is None:
        previous_visible_widget.set_visible(False)
    return visible_count


class ContextualMenu(Gtk.Menu):
    def __init__(self, main_entries):
        super().__init__()
        self.main_entries = main_entries

    def add_menuitem(self, entry):
        """Add a menu item to the current menu

        Params:
            entry (tuple): tuple containing name, label and callback

        Returns:
            Gtk.MenuItem
        """
        name, label, callback = entry
        if label == "-":
            separator = Gtk.SeparatorMenuItem()
            self.append(separator)
            return separator

        action = Gtk.Action(name=name, label=label)
        action.connect("activate", callback)

        menu_item = action.create_menu_item()
        menu_item.action_id = name
        self.append(menu_item)
        return menu_item

    def popup(self, event, game_actions):
        for item in self.get_children():
            self.remove(item)

        for entry in self.main_entries:
            self.add_menuitem(entry)

        self.show_all()

        displayed = game_actions.get_displayed_entries()

        def is_visible(w):
            if isinstance(w, Gtk.SeparatorMenuItem):
                return None

            return displayed.get(w.action_id, True)

        visible_count = update_action_widget_visibility(self.get_children(), is_visible)

        if visible_count > 0:
            self.popup_at_pointer(event)



# Standard Library
import os
from gettext import gettext as _

# Lutris Modules
from lutris import settings
from lutris.exceptions import GameConfigError, MisconfigurationError, MissingExecutableError, MissingGameExecutableError
from lutris.runners.runner import Runner
from lutris.util import system
from lutris.util.log import logger


class vice(Runner):
    description = _("Commodore Emulator")
    human_name = _("Vice")
    # flatpak_id = "net.sf.VICE"  # needs adjustments
    platforms = [
        _("Commodore 64"),
        _("Commodore 128"),
        _("Commodore VIC20"),
        _("Commodore PET"),
        _("Commodore Plus/4"),
        _("Commodore CBM II"),
    ]
    machine_choices = [
        ("C64", "c64"),
        ("C128", "c128"),
        ("vic20", "vic20"),
        ("PET", "pet"),
        ("Plus/4", "plus4"),
        ("CBM-II", "cbmii"),
    ]
    game_options = [
        {
            "option": "main_file",
            "type": "file",
            "label": _("ROM file"),
            "help": _(
                "The game data, commonly called a ROM image.\n"
                "Supported formats: X64, D64, G64, P64, D67, D71, D81, "
                "D80, D82, D1M, D2M, D4M, T46, P00 and CRT."
            ),
        }
    ]

    runner_options = [
        {"option": "joy", "type": "bool", "label": _("Use joysticks"), "default": False},
        {
            "option": "fullscreen",
            "type": "bool",
            "section": _("Graphics"),
            "label": _("Fullscreen"),
            "default": False,
        },
        {
            "option": "double",
            "type": "bool",
            "section": _("Graphics"),
            "label": _("Scale up display by 2"),
            "default": True,
        },
        {
            "option": "aspect_ratio",
            "type": "bool",
            "section": _("Graphics"),
            "label": _("Preserve aspect ratio"),
            "default": True,
        },
        {
            "option": "renderer",
            "type": "choice",
            "section": _("Graphics"),
            "label": _("Graphics renderer"),
            "choices": [("OpenGL", "opengl"), (_("Software"), "software")],
            "default": "opengl",
        },
        {
            "option": "drivesound",
            "type": "bool",
            "label": _("Enable sound emulation of disk drives"),
            "default": False,
        },
        {
            "option": "machine",
            "type": "choice",
            "label": _("Machine"),
            "choices": machine_choices,
            "default": "c64",
        },
    ]

    def get_platform(self):
        machine = self.game_config.get("machine")
        if machine:
            for index, choice in enumerate(self.machine_choices):
                if choice[1] == machine:
                    return self.platforms[index]
        return self.platforms[0]  # Default to C64

    def get_executable(self, machine: str = None) -> str:
        if not machine:
            machine = "c64"
        executables = {
            "c64": "x64",
            "c128": "x128",
            "vic20": "xvic",
            "pet": "xpet",
            "plus4": "xplus4",
            "cbmii": "xcbm2",
        }
        try:
            executable = executables[machine]
            exe = os.path.join(settings.RUNNER_DIR, "vice/bin/%s" % executable)
            if not os.path.isfile(exe):
                raise MissingExecutableError(_("The executable '%s' could not be found.") % exe)
            return exe
        except KeyError as ex:
            raise MisconfigurationError("Invalid machine '%s'" % machine) from ex

    def install(self, install_ui_delegate, version=None, callback=None):
        def on_runner_installed(*args):
            config_path = system.create_folder("~/.vice")
            lib_dir = os.path.join(settings.RUNNER_DIR, "vice/lib/vice")
            if not system.path_exists(lib_dir):
                lib_dir = os.path.join(settings.RUNNER_DIR, "vice/lib64/vice")
            if not system.path_exists(lib_dir):
                logger.error("Missing lib folder in the Vice runner")
            else:
                system.merge_folders(lib_dir, config_path)
            if callback:
                callback()

        super().install(install_ui_delegate, version, on_runner_installed)

    def get_roms_path(self, machine=None):
        if not machine:
            machine = "c64"
        paths = {
            "c64": "C64",
            "c128": "C128",
            "vic20": "VIC20",
            "pet": "PET",
            "plus4": "PLUS4",
            "cmbii": "CBM-II",
        }
        root_dir = os.path.dirname(os.path.dirname(self.get_executable()))
        return os.path.join(root_dir, "lib64/vice", paths[machine])

    @staticmethod
    def get_option_prefix(machine):
        prefixes = {
            "c64": "VICII",
            "c128": "VICII",
            "vic20": "VIC",
            "pet": "CRTC",
            "plus4": "TED",
            "cmbii": "CRTC",
        }
        return prefixes[machine]

    @staticmethod
    def get_joydevs(machine):
        joydevs = {"c64": 2, "c128": 2, "vic20": 1, "pet": 0, "plus4": 2, "cmbii": 0}
        return joydevs[machine]

    @staticmethod
    def get_rom_args(machine, rom):
        args = []

        if rom.endswith(".crt"):
            crt_option = {
                "c64": "-cartcrt",
                "c128": "-cartcrt",
                "vic20": "-cartgeneric",
                "pet": None,
                "plus4": "-cart",
                "cmbii": None,
            }
            if crt_option[machine]:
                args.append(crt_option[machine])

        args.append(rom)
        return args

    def play(self):
        machine = self.runner_config.get("machine")

        rom = self.game_config.get("main_file")
        if not rom:
            raise GameConfigError(_("No rom provided"))
        if not system.path_exists(rom):
            raise MissingGameExecutableError(filename=rom)

        params = [self.get_executable(machine)]
        rom_dir = os.path.dirname(rom)
        params.append("-chdir")
        params.append(rom_dir)
        option_prefix = self.get_option_prefix(machine)

        if self.runner_config.get("fullscreen"):
            params.append("-{}full".format(option_prefix))

        if self.runner_config.get("double"):
            params.append("-{}dsize".format(option_prefix))

        if self.runner_config.get("renderer"):
            params.append("-sdl2renderer")
            params.append(self.runner_config["renderer"])

        if not self.runner_config.get("aspect_ratio", True):
            params.append("-sdlaspectmode")
            params.append("0")

        if self.runner_config.get("drivesound"):
            params.append("-drivesound")

        if self.runner_config.get("joy"):
            for dev in range(self.get_joydevs(machine)):
                params += ["-joydev{}".format(dev + 1), "4"]

        params.extend(self.get_rom_args(machine, rom))
        return {"command": params}



"""Exception handling module"""

from collections.abc import Iterable
from gettext import gettext as _
from typing import Any, Optional


class LutrisError(Exception):
    """Base exception for Lutris related errors"""

    def __init__(self, message: str, message_markup: str = None, *args: Any, **kwargs: Any):
        super().__init__(message, *args, **kwargs)
        self.message = message
        self.message_markup = message_markup
        self.is_expected = False


class MissingRuntimeComponentError(LutrisError):
    """Raised when a Lutris component isn't found, but should have been installed."""

    def __init__(self, message: str, component_name: str, *args: Any, **kwargs: Any):
        super().__init__(message, *args, **kwargs)
        self.component_name = component_name
        self.is_expected = True


class MisconfigurationError(LutrisError):
    """Raised for incorrect configuration or installation, like incorrect
    or missing settings, missing components, that sort of thing. This has subclasses
    that are less vague."""


class DirectoryNotFoundError(MisconfigurationError):
    """Raise this error if a directory that is required is not present."""

    def __init__(self, message: str = None, directory: str = None, *args: Any, **kwargs: Any):
        if not message and directory:
            message = _("The directory {} could not be found").format(directory)
        super().__init__(message, *args, **kwargs)
        self.directory = directory


class SymlinkNotUsableError(MisconfigurationError):
    """Raise this error if a symlink that is required is not usable."""

    def __init__(self, message: str = None, link: str = None, *args: Any, **kwarg: Any):
        if not message and link:
            message = message or _("The link {} could not be used.").format(link)
        super().__init__(message, *args, **kwarg)
        self.link = link


class GameConfigError(MisconfigurationError):
    """Throw this error when the game configuration prevents the game from
    running properly."""


class MissingBiosError(GameConfigError):
    """Throw this error when the game requires a BIOS, but none is configured."""

    def __init__(self, message: str = None, *args: Any, **kwargs: Any):
        super().__init__(message or _("A bios file is required to run this game"), *args, **kwargs)


class AuthenticationError(LutrisError):
    """Raised when authentication to a service fails"""


class UnavailableGameError(LutrisError):
    """Raised when a game is unavailable from a service"""


class UnavailableLibrariesError(MisconfigurationError):
    def __init__(self, libraries: Iterable[str], arch: Optional[str] = None):
        message = _("The following {arch} libraries are required but are not installed on your system:\n{libs}").format(
            arch=arch if arch else "", libs=", ".join(libraries)
        )
        super().__init__(message)
        self.libraries = libraries


class UnavailableRunnerError(MisconfigurationError):
    """Raised when a runner is not installed or not installed fully."""


class UnspecifiedVersionError(MisconfigurationError):
    """Raised when a version number must be specified, but was not."""


class MissingExecutableError(MisconfigurationError):
    """Raised when a program can't be located."""


class MissingMediaError(LutrisError):
    """Raised when an image file could not be found."""

    def __init__(self, message: str = None, filename: str = None, *args: Any, **kwargs: Any):
        if not message and filename:
            message = _("The file {} could not be found").format(filename)

        super().__init__(message, *args, **kwargs)
        self.filename = filename


class MissingGameExecutableError(MissingExecutableError):
    """Raise when a game's executable can't be found is not specified."""

    def __init__(self, message: str = None, filename: str = None, *args: Any, **kwargs: Any):
        if not message:
            if filename:
                message = _("The file {} could not be found").format(filename)
            else:
                message = _("This game has no executable set. The install process didn't finish properly.")

        super().__init__(message, *args, **kwargs)
        self.filename = filename


class InvalidGameMoveError(LutrisError):
    """Raised when a game can't be moved as desired; we may have to just set the location."""


class EsyncLimitError(Exception):
    """Raised when the ESYNC limit is not set correctly."""

    def __init__(self, message: str = None, *args: Any, **kwargs: Any):
        if not message:
            message = _("Your ESYNC limits are not set correctly.")

        super().__init__(message, *args, **kwargs)


class FsyncUnsupportedError(Exception):
    """Raised when FSYNC is enabled, but is not supported by the kernel."""

    def __init__(self, message: str = None, *args: Any, **kwargs: Any):
        if not message:
            message = _("Your kernel is not patched for fsync. Please get a patched kernel to use fsync.")

        super().__init__(message, *args, **kwargs)


class InvalidSearchTermError(ValueError):
    def __init__(self, message: str, *args: Any, **kwargs: Any):
        super().__init__(message, *args, **kwargs)
        self.message = message



"""Commonly used dialogs"""

import builtins
import inspect
import os
import traceback
from gettext import gettext as _
from typing import Any, Callable, Dict, Optional, Type, TypeVar, Union, cast

import gi

from lutris.exceptions import LutrisError

gi.require_version("Gdk", "3.0")
gi.require_version("Gtk", "3.0")

from gi.repository import Gdk, GObject, Gtk

from lutris import api, settings
from lutris.gui.widgets.log_text_view import LogTextView
from lutris.gui.widgets.utils import get_widget_children, get_widget_window
from lutris.util import datapath
from lutris.util.jobs import schedule_at_idle
from lutris.util.log import get_log_contents, logger
from lutris.util.strings import gtk_safe


class Dialog(Gtk.Dialog):
    """A base class for dialogs that provides handling for the response signal;
    you can override its on_response() methods, but that method will record
    the response for you via 'response_type' or 'confirmed' and destory this
    dialog if it isn't NONE."""

    def __init__(
        self,
        title: str = None,
        parent: Gtk.Widget = None,
        flags: Gtk.DialogFlags = 0,
        buttons: Gtk.ButtonsType = None,
        **kwargs,
    ):
        # MyPy can't see it, but __init__ can handle the new_with_buttons arguments for us
        super().__init__(title, parent, flags, buttons, **kwargs)  # type:ignore
        self._response_type = Gtk.ResponseType.NONE
        self.connect("response", self.on_response)

    @property
    def response_type(self) -> Gtk.ResponseType:
        """The response type of the response that occurred; initially this is NONE.
        Use the GTK response() method to artificially generate a response, rather than
        setting this."""
        return self._response_type

    @property
    def confirmed(self) -> bool:
        """True if 'response_type' is OK or YES."""
        return self.response_type in (Gtk.ResponseType.OK, Gtk.ResponseType.YES)

    def on_response(self, _dialog, response: Gtk.ResponseType) -> None:
        """Handles the dialog response; you can override this but by default
        this records the response for 'response_type'."""
        self._response_type = response

    def destroy_at_idle(self, condition: Optional[Callable] = None):
        """Adds as idle task to destroy this window at idle time;
        it can do so conditionally if you provide a callable to check,
        but it checks only once. You can still explicitly destroy the
        dialog after calling this. This is used to ensure destruction of
        ModalDialog after run()."""

        def idle_destroy():
            if not condition or condition():
                self.destroy()

        def on_destroy(*_args):
            self.disconnect(on_destroy_id)
            idle_destroy_task.unschedule()

        self.hide()
        idle_destroy_task = schedule_at_idle(idle_destroy)
        on_destroy_id = self.connect("destroy", on_destroy)

    def add_styled_button(self, button_text: str, response_id: Gtk.ResponseType, css_class: str):
        button = self.add_button(button_text, response_id)
        if css_class:
            style_context = button.get_style_context()
            style_context.add_class(css_class)
        return button

    def add_default_button(self, button_text: str, response_id: Gtk.ResponseType, css_class: str = "suggested-action"):
        """Adds a button to the dialog with a particular response id, but
        also makes it the default and styles it as the suggested action."""
        button = self.add_styled_button(button_text, response_id, css_class)
        self.set_default_response(response_id)
        return button


class ModalDialog(Dialog):
    """A base class of modal dialogs, which sets the flag for you.

    Unlike plain Gtk.Dialog, these destroy themselves (at idle-time) after
    you call run(), even if you forget to. They aren't meant to be reused."""

    def __init__(
        self,
        title: str = None,
        parent: Gtk.Widget = None,
        flags: Gtk.DialogFlags = 0,
        buttons: Gtk.ButtonsType = None,
        **kwargs,
    ):
        super().__init__(title, parent, flags | Gtk.DialogFlags.MODAL, buttons, **kwargs)
        self.set_destroy_with_parent(True)

    def on_response(self, dialog, response: Gtk.ResponseType) -> None:
        super().on_response(dialog, response)
        # Model dialogs do return from run() in response from respose() but the
        # dialog is visible and locks out its parent. So we hide it. Watch out-
        # self.destroy() changes the run() result to NONE.
        if response != Gtk.ResponseType.NONE:
            self.hide()
            self.destroy_at_idle(condition=lambda: not self.get_visible())


class ModelessDialog(Dialog):
    """A base class for modeless dialogs. They have a parent only temporarily, so
    they can be centered over it during creation. But each modeless dialog gets
    its own window group, so it treats its own modal dialogs separately, and it resets
    its transient-for after being created."""

    def __init__(
        self,
        title: str = None,
        parent: Gtk.Widget = None,
        flags: Gtk.DialogFlags = 0,
        buttons: Gtk.ButtonsType = None,
        **kwargs,
    ):
        super().__init__(title, parent, flags, buttons, **kwargs)
        # These are not stuck above the 'main' window, but can be
        # re-ordered freely.
        self.set_type_hint(Gdk.WindowTypeHint.NORMAL)

        # These are independent windows, but start centered over
        # a parent like a dialog. Not modal, not really transient,
        # and does not share modality with other windows - so it
        # needs its own window group.
        Gtk.WindowGroup().add_window(self)
        schedule_at_idle(self._clear_transient_for)

    def _clear_transient_for(self) -> None:
        # we need the parent set to be centered over the parent, but
        # we don't want to be transient really - we want other windows
        # able to come to the front.
        self.set_transient_for(None)

    def on_response(self, dialog, response: Gtk.ResponseType) -> None:
        super().on_response(dialog, response)
        # Modal dialogs self-destruct, but modeless ones must commit
        # suicide more explicitly.
        if response != Gtk.ResponseType.NONE:
            self.destroy()


class SavableModelessDialog(ModelessDialog):
    """This is a modeless dialog that has a Cancel and a Save button in the header-bar,
    with a ctrl-S keyboard shortcut to save."""

    def __init__(self, title: str, parent: Gtk.Widget = None, **kwargs):
        super().__init__(title, parent=parent, use_header_bar=True, **kwargs)

        self.cancel_button = self.add_button(_("Cancel"), Gtk.ResponseType.CANCEL)
        self.cancel_button.set_valign(Gtk.Align.CENTER)

        self.save_button = self.add_styled_button(_("Save"), Gtk.ResponseType.NONE, css_class="suggested-action")
        self.save_button.set_valign(Gtk.Align.CENTER)
        self.save_button.connect("clicked", self.on_save)

        self.accelerators = Gtk.AccelGroup()
        self.add_accel_group(self.accelerators)
        key, mod = Gtk.accelerator_parse("<Primary>s")
        self.save_button.add_accelerator("clicked", self.accelerators, key, mod, Gtk.AccelFlags.VISIBLE)

    def on_save(self, _button):
        pass


class GtkBuilderDialog(GObject.Object):
    dialog_object = NotImplemented

    __gsignals__ = {
        "destroy": (GObject.SignalFlags.RUN_LAST, None, ()),
    }

    def __init__(self, parent=None, **kwargs):
        # pylint: disable=no-member
        super().__init__()
        ui_filename = os.path.join(datapath.get(), "ui", self.glade_file)
        if not os.path.exists(ui_filename):
            raise ValueError("ui file does not exists: %s" % ui_filename)

        self.builder = Gtk.Builder()
        self.builder.add_from_file(ui_filename)
        self.dialog = self.builder.get_object(self.dialog_object)

        self.builder.connect_signals(self)
        if parent:
            self.dialog.set_transient_for(parent)
        self.dialog.show_all()
        self.dialog.connect("delete-event", self.on_close)
        self.initialize(**kwargs)

    def initialize(self, **kwargs):
        """Implement further customizations in subclasses"""

    def present(self):
        self.dialog.present()

    def on_close(self, *args):  # pylint: disable=unused-argument
        """Propagate the destroy event after closing the dialog"""
        self.dialog.destroy()
        self.emit("destroy")

    def on_response(self, widget, response):  # pylint: disable=unused-argument
        if response == Gtk.ResponseType.DELETE_EVENT:
            try:
                self.dialog.hide()
            except AttributeError:
                pass


class AboutDialog(GtkBuilderDialog):
    glade_file = "about-dialog.ui"
    dialog_object = "about_dialog"

    def initialize(self):  # pylint: disable=arguments-differ
        self.dialog.set_version(settings.VERSION)


class NoticeDialog(Gtk.MessageDialog):
    """Display a message to the user."""

    def __init__(self, message_markup: str, secondary: Optional[str] = None, parent: Optional[Gtk.Widget] = None):
        parent: Gtk.Window = get_widget_window(parent)
        super().__init__(message_type=Gtk.MessageType.INFO, buttons=Gtk.ButtonsType.OK, parent=parent)
        self.set_markup(message_markup)
        if secondary:
            self.format_secondary_text(secondary[:256])

        # So you can copy warning text
        for child in get_widget_children(self.get_message_area(), child_type=Gtk.Label):
            child.set_selectable(True)

        self.run()
        self.destroy()


class WarningDialog(Gtk.MessageDialog):
    """Display a warning to the user, who responds with whether to proceed, like
    a QuestionDialog."""

    def __init__(self, message_markup: str, secondary: Optional[str] = None, parent: Optional[Gtk.Widget] = None):
        parent: Gtk.Window = get_widget_window(parent)
        super().__init__(message_type=Gtk.MessageType.WARNING, buttons=Gtk.ButtonsType.OK_CANCEL, parent=parent)
        self.set_markup(message_markup)
        if secondary:
            self.format_secondary_text(secondary[:256])

        # So you can copy warning text
        for child in get_widget_children(self.get_message_area(), child_type=Gtk.Label):
            child.set_selectable(True)

        self.result = self.run()
        self.destroy()


class ErrorDialog(Gtk.MessageDialog):
    """Display an error message."""

    def __init__(
        self,
        error: Union[str, builtins.BaseException],
        message_markup: Optional[str] = None,
        secondary_markup: Optional[str] = None,
        parent: Optional[Gtk.Widget] = None,
    ):
        parent: Gtk.Window = get_widget_window(parent)
        super().__init__(message_type=Gtk.MessageType.ERROR, buttons=Gtk.ButtonsType.OK, parent=parent)

        def get_message_markup(err: Union[BaseException, str]) -> str:
            if isinstance(err, LutrisError):
                return err.message_markup or gtk_safe(str(err))
            else:
                return gtk_safe(str(err))

        if isinstance(error, builtins.BaseException):
            if secondary_markup:
                # Some errors contain < and > and look like markup, but aren't-
                # we'll need to protect the message dialog against this. To use markup,
                # you must pass the message itself directly.
                message_markup = message_markup or get_message_markup(error)
            elif not message_markup:
                message_markup = "<span weight='bold'>%s</span>" % _("Lutris has encountered an error")
                secondary_markup = get_message_markup(error)
        elif not message_markup:
            message_markup = get_message_markup(error)

        # Gtk doesn't wrap long labels containing no space correctly
        # the length of the message is limited to avoid display issues

        if message_markup:
            self.set_markup(message_markup[:256])

        if secondary_markup:
            self.format_secondary_markup(secondary_markup[:256])

        # So you can copy error text
        for child in get_widget_children(self.get_message_area(), child_type=Gtk.Label):
            child.set_selectable(True)

        if isinstance(error, BaseException):
            content_area = self.get_content_area()
            spacing = content_area.get_spacing()
            content_area.set_spacing(0)

            details_expander = self.get_details_expander(error)
            details_expander.set_margin_top(spacing)
            content_area.pack_end(details_expander, False, False, 0)

            action_area = cast(Gtk.ButtonBox, self.get_action_area())
            copy_button = Gtk.Button(label=_("Copy Details to Clipboard"), visible=True)
            action_area.pack_start(copy_button, False, True, 0)
            action_area.set_child_secondary(copy_button, True)
            copy_button.connect("clicked", self.on_copy_clicked, error)

        self.run()
        self.destroy()

    def on_copy_clicked(self, _button, error: BaseException):
        details = self.format_error(error)
        clipboard = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)
        clipboard.set_text(details, -1)

    def get_details_expander(self, error: BaseException) -> Gtk.Widget:
        details = self.format_error(error, include_message=False)

        box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        label = Gtk.Label(xalign=0.0, wrap=True, margin_left=6, margin_right=6, margin_bottom=6)
        label.set_markup(
            _(
                "You can get support from "
                "<a href='https://github.com/lutris/lutris'>GitHub</a> or "
                "<a href='https://discordapp.com/invite/Pnt5CuY'>Discord</a>. "
                "Make sure to provide the error details;\n"
                "use the 'Copy Details to Clipboard' button to get them."
            )
        )
        box.pack_start(label, False, False, 0)

        expander = Gtk.Expander.new(_("Error details"))

        details_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)

        details_box.pack_start(Gtk.Separator(), False, False, 0)

        details_textview = Gtk.TextView(editable=False)
        details_textview.get_buffer().set_text(details)

        details_scrolledwindow = Gtk.ScrolledWindow(width_request=800, height_request=400)
        details_scrolledwindow.add(details_textview)
        details_box.pack_start(details_scrolledwindow, False, False, 0)
        expander.add(details_box)

        box.pack_start(expander, True, True, 0)
        box.show_all()
        return box

    @staticmethod
    def format_error(error: BaseException, include_message: bool = True):
        formatted = traceback.format_exception(type(error), error, error.__traceback__)
        if include_message:
            formatted = [str(error), ""] + formatted
        text = "\n".join(formatted).strip()
        log = get_log_contents()

        if log:
            text = f"{text}\n\nLutris log:\n{log}".strip()

        return text


class QuestionDialog(Gtk.MessageDialog):
    """Ask the user a yes or no question."""

    YES = Gtk.ResponseType.YES
    NO = Gtk.ResponseType.NO

    def __init__(self, dialog_settings):
        super().__init__(message_type=Gtk.MessageType.QUESTION, buttons=Gtk.ButtonsType.YES_NO)
        self.set_markup(dialog_settings["question"])
        self.set_title(dialog_settings["title"])
        if "parent" in dialog_settings:
            self.set_transient_for(dialog_settings["parent"])
        if "widgets" in dialog_settings:
            for widget in dialog_settings["widgets"]:
                self.get_message_area().add(widget)
        self.result = self.run()
        self.destroy()


class InputDialog(ModalDialog):
    """Ask the user for a text input"""

    def __init__(self, dialog_settings):
        super().__init__(parent=dialog_settings["parent"])
        self.set_border_width(12)
        self.user_value = ""
        self.add_button(Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL)
        self.ok_button = self.add_default_button(Gtk.STOCK_OK, Gtk.ResponseType.OK)
        self.set_default_response(Gtk.ResponseType.OK)
        self.ok_button.set_sensitive(False)
        self.set_title(dialog_settings["title"])
        label = Gtk.Label(visible=True)
        label.set_markup(dialog_settings["question"])
        self.get_content_area().pack_start(label, True, True, 12)
        self.entry = Gtk.Entry(visible=True, activates_default=True)
        self.entry.connect("changed", self.on_entry_changed)
        self.get_content_area().pack_start(self.entry, True, True, 12)
        self.entry.set_text(dialog_settings.get("initial_value") or "")

    def on_entry_changed(self, widget):
        self.user_value = widget.get_text()
        self.ok_button.set_sensitive(bool(self.user_value))


class DirectoryDialog:
    """Ask the user to select a directory."""

    def __init__(self, message, default_path=None, parent=None):
        self.folder = None
        dialog = Gtk.FileChooserNative.new(
            message,
            parent,
            Gtk.FileChooserAction.SELECT_FOLDER,
            _("_OK"),
            _("_Cancel"),
        )
        if default_path:
            dialog.set_current_folder(default_path)
        self.result = dialog.run()
        if self.result == Gtk.ResponseType.ACCEPT:
            self.folder = dialog.get_filename()
        dialog.destroy()


class FileDialog:
    """Ask the user to select a file."""

    def __init__(self, message=None, default_path=None, mode="open", parent=None):
        self.filename = None
        if not message:
            message = _("Please choose a file")
        if mode == "save":
            action = Gtk.FileChooserAction.SAVE
        else:
            action = Gtk.FileChooserAction.OPEN
        dialog = Gtk.FileChooserNative.new(
            message,
            parent,
            action,
            _("_OK"),
            _("_Cancel"),
        )
        if default_path and os.path.exists(default_path):
            dialog.set_current_folder(default_path)
        dialog.set_local_only(False)
        response = dialog.run()
        if response == Gtk.ResponseType.ACCEPT:
            self.filename = dialog.get_filename()

        dialog.destroy()


class InstallOrPlayDialog(ModalDialog):
    def __init__(self, game_name, parent=None):
        super().__init__(title=_("%s is already installed") % game_name, parent=parent, border_width=10)
        self.action = "play"
        self.action_confirmed = False

        self.add_button(Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL)
        self.add_default_button(Gtk.STOCK_OK, Gtk.ResponseType.OK)

        self.set_size_request(320, 120)
        vbox = Gtk.Box.new(Gtk.Orientation.VERTICAL, 6)
        self.get_content_area().add(vbox)
        play_button = Gtk.RadioButton.new_with_label_from_widget(None, _("Launch game"))
        play_button.connect("toggled", self.on_button_toggled, "play")
        vbox.pack_start(play_button, False, False, 0)
        install_button = Gtk.RadioButton.new_from_widget(play_button)
        install_button.set_label(_("Install the game again"))
        install_button.connect("toggled", self.on_button_toggled, "install")
        vbox.pack_start(install_button, False, False, 0)

        self.show_all()
        self.run()

    def on_button_toggled(self, _button, action):
        logger.debug("Action set to %s", action)
        self.action = action

    def on_response(self, _widget, response):
        if response == Gtk.ResponseType.CANCEL:
            self.action = None
        super().on_response(_widget, response)


class LaunchConfigSelectDialog(ModalDialog):
    def __init__(self, game, configs, title, parent=None):
        super().__init__(title=title, parent=parent, border_width=10)
        self.config_index = 0
        self.dont_show_again = False

        self.add_button(Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL)
        self.add_default_button(Gtk.STOCK_OK, Gtk.ResponseType.OK)

        self.set_size_request(320, 120)
        vbox = Gtk.Box.new(Gtk.Orientation.VERTICAL, 6)
        self.get_content_area().add(vbox)

        primary_game_radio = Gtk.RadioButton.new_with_label_from_widget(None, game.name)
        primary_game_radio.connect("toggled", self.on_button_toggled, 0)
        vbox.pack_start(primary_game_radio, False, False, 0)
        for i, config in enumerate(configs):
            _button = Gtk.RadioButton.new_from_widget(primary_game_radio)
            _button.set_label(config["name"])
            _button.connect("toggled", self.on_button_toggled, i + 1)
            vbox.pack_start(_button, False, False, 0)

        dont_show_checkbutton = Gtk.CheckButton(_("Do not ask again for this game."))
        dont_show_checkbutton.connect("toggled", self.on_dont_show_checkbutton_toggled)
        vbox.pack_end(dont_show_checkbutton, False, False, 6)

        self.show_all()
        self.run()

    def on_button_toggled(self, _button, index):
        self.config_index = index

    def on_dont_show_checkbutton_toggled(self, _button):
        self.dont_show_again = _button.get_active()


class ClientLoginDialog(GtkBuilderDialog):
    glade_file = "dialog-lutris-login.ui"
    dialog_object = "lutris-login"
    __gsignals__ = {
        "connected": (GObject.SignalFlags.RUN_LAST, None, (GObject.TYPE_PYOBJECT,)),
        "cancel": (GObject.SignalFlags.RUN_LAST, None, (GObject.TYPE_PYOBJECT,)),
    }

    def __init__(self, parent):
        super().__init__(parent=parent)

        self.parent = parent
        self.username_entry = self.builder.get_object("username_entry")
        self.password_entry = self.builder.get_object("password_entry")

        cancel_button = self.builder.get_object("cancel_button")
        cancel_button.connect("clicked", self.on_close)
        connect_button = self.builder.get_object("connect_button")
        connect_button.connect("clicked", self.on_connect)

    def get_credentials(self):
        username = self.username_entry.get_text()
        password = self.password_entry.get_text()
        return username, password

    def on_username_entry_activate(self, widget):  # pylint: disable=unused-argument
        if all(self.get_credentials()):
            self.on_connect(None)
        else:
            self.password_entry.grab_focus()

    def on_password_entry_activate(self, widget):  # pylint: disable=unused-argument
        if all(self.get_credentials()):
            self.on_connect(None)
        else:
            self.username_entry.grab_focus()

    def on_connect(self, widget):  # pylint: disable=unused-argument
        username, password = self.get_credentials()
        token = api.connect(username, password)
        if not token:
            NoticeDialog(_("Login failed"), parent=self.parent)
        else:
            self.dialog.destroy()
            self.emit("connected", username)


class InstallerSourceDialog(ModelessDialog):
    """Show install script source"""

    def __init__(self, code, name, parent):
        super().__init__(title=_("Install script for {}").format(name), parent=parent, border_width=0)
        self.set_default_size(800, 750)

        ok_button = self.add_default_button(Gtk.STOCK_OK, Gtk.ResponseType.OK)
        ok_button.set_border_width(10)

        self.scrolled_window = Gtk.ScrolledWindow()
        self.scrolled_window.set_hexpand(True)
        self.scrolled_window.set_vexpand(True)

        source_buffer = Gtk.TextBuffer()
        source_buffer.set_text(code)

        source_box = LogTextView(source_buffer, autoscroll=False)

        self.get_content_area().set_border_width(0)
        self.get_content_area().add(self.scrolled_window)
        self.scrolled_window.add(source_box)

        self.show_all()


class HumbleBundleCookiesDialog(ModalDialog):
    def __init__(self, parent=None):
        super().__init__(_("Humble Bundle Cookie Authentication"), parent)
        self.cookies_content = None
        self.add_button(Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL)
        self.add_default_button(Gtk.STOCK_OK, Gtk.ResponseType.OK)

        self.set_size_request(640, 512)

        vbox = Gtk.Box.new(Gtk.Orientation.VERTICAL, 6)
        self.get_content_area().add(vbox)
        label = Gtk.Label()
        label.set_markup(
            _(
                "<b>Humble Bundle Authentication via cookie import</b>\n"
                "\n"
                "<b>In Firefox</b>\n"
                "- Install the following extension: "
                "<a href='https://addons.mozilla.org/en-US/firefox/addon/export-cookies-txt/'>"
                "https://addons.mozilla.org/en-US/firefox/addon/export-cookies-txt/"
                "</a>\n"
                "- Open a tab to humblebundle.com and make sure you are logged in.\n"
                "- Click the cookie icon in the top right corner, next to the settings menu\n"
                "- Check 'Prefix HttpOnly cookies' and click 'humblebundle.com'\n"
                "- Open the generated file and paste the contents below. Click OK to finish.\n"
                "- You can delete the cookies file generated by Firefox\n"
                "- Optionally, <a href='https://support.humblebundle.com/hc/en-us/requests/new'>"
                "open a support ticket</a> to ask Humble Bundle to fix their configuration."
            )
        )
        vbox.pack_start(label, False, False, 24)
        self.textview = Gtk.TextView()
        self.textview.set_left_margin(12)
        self.textview.set_right_margin(12)
        scrolledwindow = Gtk.ScrolledWindow()
        scrolledwindow.set_hexpand(True)
        scrolledwindow.set_vexpand(True)
        scrolledwindow.add(self.textview)
        vbox.pack_start(scrolledwindow, True, True, 24)
        self.show_all()
        self.run()

    def on_response(self, dialog, response):
        if response == Gtk.ResponseType.CANCEL:
            self.cookies_content = None
        else:
            buffer = self.textview.get_buffer()
            self.cookies_content = buffer.get_text(buffer.get_start_iter(), buffer.get_end_iter(), True)

        super().on_response(dialog, response)


def _call_when_destroyed(self: Gtk.Widget, callback: Callable[[], None]) -> Callable[[], None]:
    handler_id = self.connect("destroy", lambda *x: callback())
    return lambda: self.disconnect(handler_id)


# call_when_destroyed is a utility that hooks up the 'destroy' signal to call your callback,
# and returns a callable that unhooks it. This is used by AsyncJob to avoid sending a callback
# to a destroyed widget.
Gtk.Widget.call_when_destroyed = _call_when_destroyed  # type: ignore[attr-defined]

_error_handlers: Dict[Type[BaseException], Callable[[BaseException, Gtk.Window], Any]] = {}
TError = TypeVar("TError", bound=BaseException)


def display_error(error: BaseException, parent: Gtk.Widget) -> None:
    """Displays an error in a modal dialog. This can be customized via
    register_error_handler(), but displays an ErrorDialog by default.

    This allows custom error handling to be invoked anywhere that can show an
    ErrorDialog, instead of having to bounce exceptions off the backstop."""
    handler = get_error_handler(type(error))

    if isinstance(parent, Gtk.Window):
        handler(error, parent)
    else:
        handler(error, cast(Gtk.Window, parent.get_toplevel()))


def register_error_handler(error_class: Type[TError], handler: Callable[[TError, Gtk.Window], Any]) -> None:
    """Records a function to call to handle errors of a particular class or its subclasses. The
    function is given the error and a parent window, and can display a modal dialog."""
    _error_handlers[error_class] = handler


def get_error_handler(error_class: Type[TError]) -> Callable[[TError, Gtk.Window], Any]:
    """Returns the register error handler for an exception class. If none is registered,
    this returns a default handler that shows an ErrorDialog."""
    if not isinstance(error_class, type):
        if isinstance(error_class, BaseException):
            logger.debug("An error was passed where an error class should be passed.")
            error_class = type(error_class)
        else:
            raise ValueError(f"'{error_class}' was passed to get_error_handler, but an error class is required here.")

    if error_class in _error_handlers:
        return _error_handlers[error_class]

    for base_class in inspect.getmro(error_class):
        if base_class in _error_handlers:
            return _error_handlers[base_class]

    return lambda e, p: ErrorDialog(e, parent=p)


def _handle_keyerror(error: KeyError, parent: Gtk.Window) -> None:
    message = _("The key '%s' could not be found.") % error.args[0]
    ErrorDialog(error, message_markup=gtk_safe(message), parent=parent)


register_error_handler(KeyError, _handle_keyerror)



"""Common values used for views"""

(
    COL_ID,
    COL_SLUG,
    COL_NAME,
    COL_SORTNAME,
    COL_MEDIA_PATHS,
    COL_YEAR,
    COL_RUNNER,
    COL_RUNNER_HUMAN_NAME,
    COL_PLATFORM,
    COL_LASTPLAYED,
    COL_LASTPLAYED_TEXT,
    COL_INSTALLED,
    COL_INSTALLED_AT,
    COL_INSTALLED_AT_TEXT,
    COL_PLAYTIME,
    COL_PLAYTIME_TEXT,
) = list(range(16))

COLUMN_NAMES = {
    COL_NAME: "name",
    COL_YEAR: "year",
    COL_RUNNER_HUMAN_NAME: "runner",
    COL_PLATFORM: "platform",
    COL_LASTPLAYED_TEXT: "lastplayed",
    COL_INSTALLED_AT_TEXT: "installedat",
    COL_PLAYTIME_TEXT: "playtime",
}
